import copy
import numpy as np
from yggdrasil import backwards
from yggdrasil.tests import assert_raises
from yggdrasil.serialize import AsciiTableSerialize
from yggdrasil.serialize.tests import test_DefaultSerialize as parent


def test_serialize_nofmt():
    r"""Test error on serialization without a format."""
    inst = AsciiTableSerialize.AsciiTableSerialize()
    inst._initialized = True
    test_msg = np.zeros((5, 5))
    assert_raises(RuntimeError, inst.serialize, test_msg)

    
def test_deserialize_nofmt():
    r"""Test error on deserialization without a format."""
    inst = AsciiTableSerialize.AsciiTableSerialize()
    test_msg = b'lskdbjs;kfbj'
    test_msg = inst.encoded_datatype.serialize(test_msg, metadata={})
    assert_raises(RuntimeError, inst.deserialize, test_msg)


class TestAsciiTableSerialize(parent.TestDefaultSerialize):
    r"""Test class for AsciiTableSerialize class."""

    _cls = 'AsciiTableSerialize'
    attr_list = (copy.deepcopy(parent.TestDefaultSerialize.attr_list),
                 ['format_str', 'field_names', 'field_units', 'as_array'])
    
    def test_field_specs(self):
        r"""Test field specifiers."""
        super(TestAsciiTableSerialize, self).test_field_specs()
        # Specific to this class
        self.assert_equal(self.instance.format_str,
                          backwards.as_bytes(
                              self.testing_options['kwargs']['format_str']))
        field_names = self.testing_options['kwargs'].get('field_names', None)
        if field_names is not None:
            field_names = [backwards.as_str(x) for x in field_names]
        self.assert_equal(self.instance.field_names, field_names)
        field_units = self.testing_options['kwargs'].get('field_units', None)
        if field_units is not None:
            field_units = [backwards.as_str(x) for x in field_units]
        self.assert_equal(self.instance.field_units, field_units)


class TestAsciiTableSerializeSingle(parent.TestDefaultSerialize):
    r"""Test class for AsciiTableSerialize class."""

    _cls = 'AsciiTableSerialize'
    _empty_obj = []
    _objects = [(1, )]

    def __init__(self, *args, **kwargs):
        super(TestAsciiTableSerializeSingle, self).__init__(*args, **kwargs)
        self._inst_kwargs['format_str'] = '%d\n'

    def get_options(self):
        r"""Get testing options."""
        out = {'kwargs': {'format_str': '%d\n'},
               'empty': [],
               'objects': [(1, )],
               'extra_kwargs': {},
               'typedef': {'type': 'array',
                           'items': [{'type': 'int', 'precision': 32}]},
               'dtype': None,
               'is_user_defined': False}
        return out


class TestAsciiTableSerialize_asarray(TestAsciiTableSerialize):
    r"""Test class for AsciiTableSerialize class with as_array."""

    testing_option_kws = {'array_columns': True}



from yggdrasil.communication.FileComm import FileComm


class PlyFileComm(FileComm):
    r"""Class for handling I/O from/to a .ply file on disk.

    Args:
        name (str): The environment variable where communication address is
            stored.
        **kwargs: Additional keywords arguments are passed to parent class.

    """

    _filetype = 'ply'
    _schema_subtype_description = (
        'The file is in the `Ply <http://paulbourke.net/dataformats/ply/>`_ '
        'data format for 3D structures.')
    _default_serializer = 'ply'
    _default_extension = '.ply'



"""Tools for accessing examples from python."""
import os


ex_dict = {'gs_lesson1': ('python', 'matlab', 'c', 'cpp'),
           'gs_lesson2': ('python', 'matlab', 'c', 'cpp'),
           'gs_lesson3': ('python', 'matlab', 'c', 'cpp'),
           'gs_lesson4': ('python', 'matlab', 'c', 'cpp', 'make', 'cmake'),
           'gs_lesson4b': ('python', 'matlab', 'c', 'cpp'),
           'backwards': ('python', 'matlab', 'c', 'cpp', 'make', 'cmake'),
           'formatted_io1': ('python', 'matlab', 'c', 'cpp'),
           'formatted_io2': ('python', 'matlab', 'c', 'cpp'),
           'formatted_io3': ('python', 'matlab', 'c', 'cpp'),
           'formatted_io4': ('python', 'matlab', 'c', 'cpp'),
           'formatted_io5': ('python', 'matlab', 'c', 'cpp'),
           'formatted_io6': ('python', 'matlab', 'c', 'cpp'),
           'formatted_io7': ('python', 'matlab'),  # 'c', 'cpp'),
           'formatted_io8': ('python', 'matlab'),  # 'c', 'cpp'),
           'formatted_io9': ('python', 'matlab'),  # 'c', 'cpp'),
           'rpc_lesson1': ('python', 'matlab', 'c', 'cpp'),
           'rpc_lesson2': ('python', 'matlab', 'c', 'cpp'),
           'hello': ('python', 'matlab', 'c', 'cpp'),
           'model_error': ('python', 'matlab', 'c', 'cpp'),
           'SaM': ('python', 'matlab', 'c', 'cpp', 'all', 'all_nomatlab'),
           'ascii_io': ('python', 'matlab', 'c', 'cpp', 'all', 'all_nomatlab'),
           'rpcFib': ('python', 'matlab', 'c', 'cpp', 'all', 'all_nomatlab'),
           'maxMsg': ('python', 'matlab', 'c', 'cpp', 'all', 'all_nomatlab'),
           'timed_pipe': ('python', 'matlab', 'c', 'cpp'),
           'fakeplant': ('python', 'matlab', 'c', 'cpp', 'all', 'all_nomatlab'),
           'root_to_shoot': ('python', 'c', 'all', 'all_nomatlab')}
# TODO: This can be generated from the drivers
ext_map = {'python': '.py',
           'matlab': '.m',
           'c': '.c',
           'cpp': '.cpp',
           'executable': '',
           'make': '.cpp',
           'cmake': '.cpp'}
_example_dir = os.path.dirname(__file__)


yamls = {}
source = {}
for k, lang in ex_dict.items():
    yamls[k] = {}
    source[k] = {}
    idir = os.path.join(_example_dir, k)
    isrcdir = os.path.join(idir, 'src')
    for ilang in lang:
        # Get list of yaml & source files
        if k == 'rpcFib':
            if ilang == 'all':
                cli_l = 'python'
                par_l = 'matlab'
                srv_l = 'c'
            elif ilang == 'all_nomatlab':
                cli_l = 'python'
                par_l = 'cpp'
                srv_l = 'c'
            else:
                cli_l = ilang
                par_l = ilang
                srv_l = ilang
            yml_names = ['%sCli_%s.yml' % (k, cli_l),
                         '%sCliPar_%s.yml' % (k, par_l),
                         '%sSrv_%s.yml' % (k, srv_l)]
            src_names = ['%sCli%s' % (k, ext_map[cli_l]),
                         '%sCliPar%s' % (k, ext_map[par_l]),
                         '%sSrv%s' % (k, ext_map[srv_l])]
        elif k == 'maxMsg':
            if ilang == 'all':
                cli_l = 'python'
                srv_l = 'matlab'
            elif ilang == 'all_nomatlab':
                cli_l = 'python'
                srv_l = 'c'
            else:
                cli_l = ilang
                srv_l = ilang
            yml_names = ['%sCli_%s.yml' % (k, cli_l),
                         '%sSrv_%s.yml' % (k, srv_l)]
            src_names = ['%s%s' % (k, ext_map[cli_l]),
                         '%s%s' % (k, ext_map[srv_l])]
        elif k in ['gs_lesson4', 'gs_lesson4b', 'backwards',
                   'formatted_io1', 'formatted_io2', 'formatted_io3',
                   'formatted_io4', 'formatted_io5', 'formatted_io6',
                   'formatted_io7', 'formatted_io8', 'formatted_io9']:
            yml_names = ['%s_%s.yml' % (k, ilang)]
            src_names = ['%s_modelA%s' % (k, ext_map[ilang]),
                         '%s_modelB%s' % (k, ext_map[ilang])]
        elif k in ['rpc_lesson1', 'rpc_lesson2']:
            yml_names = ['server_python.yml',
                         'client_%s.yml' % ilang]
            src_names = ['server.py', 'client%s' % ext_map[ilang]]
        elif k == 'root_to_shoot':
            if ilang.startswith('all'):
                yml_names = ['root.yml', 'shoot.yml', 'root_to_shoot.yml']
                src_names = ['root.c', 'shoot.py']
            elif ilang == 'python':
                yml_names = ['shoot.yml', 'shoot_files.yml']
                src_names = ['shoot.py']
            elif ilang == 'c':
                yml_names = ['root.yml', 'root_files.yml']
                src_names = ['root.c']
        elif k == 'fakeplant':
            if ilang.startswith('all'):
                yml_names = ['canopy.yml', 'light.yml', 'photosynthesis.yml',
                             'fakeplant.yml']
                src_names = ['canopy.cpp', 'light.c', 'photosynthesis.py']
                if ilang == 'all_nomatlab':
                    yml_names.append('growth_python.yml')
                    src_names.append('growth.py')
                else:
                    yml_names.append('growth.yml')
                    src_names.append('growth.m')
            elif ilang == 'python':
                yml_names = ['photosynthesis.yml', 'photosynthesis_files.yml']
                src_names = ['photosynthesis.py']
            elif ilang == 'c':
                yml_names = ['light.yml', 'light_files.yml']
                src_names = ['light.c']
            elif ilang == 'cpp':
                yml_names = ['canopy.yml', 'canopy_files.yml']
                src_names = ['canopy.cpp']
            elif ilang == 'matlab':
                yml_names = ['growth.yml', 'growth_files.yml']
                src_names = ['growth.m']
        else:
            yml_names = ['%s_%s.yml' % (k, ilang)]
            if ilang.startswith('all'):
                src_names = []
                for lsrc in lang:
                    if lsrc.startswith('all'):
                        continue
                    if ilang == 'all_nomatlab' and lsrc == 'matlab':
                        continue
                    src_names.append('%s%s' % (k, ext_map[lsrc]))
            else:
                src_names = ['%s%s' % (k, ext_map[ilang])]
        # Add full path to yaml & source files
        yamls[k][ilang] = [os.path.join(idir, y) for y in yml_names]
        source[k][ilang] = [os.path.join(isrcdir, s) for s in src_names]
        if len(yamls[k][ilang]) == 1:
            yamls[k][ilang] = yamls[k][ilang][0]
        if len(source[k][ilang]) == 1:
            source[k][ilang] = source[k][ilang][0]
                                   
              
__all__ = ['yamls', 'source']



from yggdrasil import languages, tools
from yggdrasil.tests import assert_raises


def test_get_language_dir():
    r"""Test the get_language_dir method."""
    assert_raises(ValueError, languages.get_language_dir, 'invalid')
    test_lang = tools.get_supported_lang()
    test_skip = ['make', 'cmake', 'executable']
    for lang in test_lang:
        if lang in test_skip:
            continue
        languages.get_language_dir(lang)



from yggdrasil.serialize.tests import test_SerializeBase as parent


class TestMatSerialize(parent.TestSerializeBase):
    r"""Test class for TestMatSerialize class."""

    _cls = 'MatSerialize'

    def test_serialize_errors(self):
        r"""Test serialize errors."""
        self.assert_raises(TypeError, self.instance.serialize, ['blah', 'blah'])



import unittest
import uuid
import copy
from yggdrasil.communication import new_comm
from yggdrasil.communication.tests import test_CommBase


class TestServerComm(test_CommBase.TestCommBase):
    r"""Tests for ServerComm communication class."""

    comm = 'ServerComm'
    attr_list = (copy.deepcopy(test_CommBase.TestCommBase.attr_list)
                 + ['response_kwargs', 'icomm', 'ocomm'])

    @property
    def send_inst_kwargs(self):
        r"""dict: Keyword arguments for send instance."""
        return {'comm': 'ClientComm'}
    
    @unittest.skipIf(True, 'Server')
    def test_error_send(self):
        r"""Disabled: Test error on send."""
        pass  # pragma: no cover
        
    @unittest.skipIf(True, 'Server')
    def test_error_recv(self):
        r"""Disabled: Test error on recv."""
        pass  # pragma: no cover
        
    @unittest.skipIf(True, 'Server')
    def test_invalid_direction(self):
        r"""Disabled: Test of error on incorrect direction."""
        pass  # pragma: no cover
    
    @unittest.skipIf(True, 'Server')
    def test_work_comm(self):
        r"""Disabled: Test creating/removing a work comm."""
        pass  # pragma: no cover

    def test_newcomm_server(self):
        r"""Test creation of server using newcomm."""
        inst = new_comm('testserver_%s' % str(uuid.uuid4()), comm=self.comm)
        self.remove_instance(inst)
        
    def test_eof_no_close(self):
        r"""Test send/recv of EOF message with no close."""
        # Forwards
        self.recv_instance.icomm.close_on_eof_recv = False
        self.do_send_recv(send_meth='send_eof', close_on_recv_eof=False)

    def test_call(self):
        r"""Test RPC call."""
        self.send_instance.sched_task(0.0, self.send_instance.rpcCall,
                                      args=[self.test_msg], store_output=True)
        flag, msg_recv = self.recv_instance.rpcRecv(timeout=self.timeout)
        assert(flag)
        self.assert_equal(msg_recv, self.test_msg)
        flag = self.recv_instance.rpcSend(msg_recv)
        assert(flag)
        T = self.recv_instance.start_timeout()
        while (not T.is_out) and (self.send_instance.sched_out is None):  # pragma: debug
            self.recv_instance.sleep()
        self.recv_instance.stop_timeout()
        flag, msg_recv = self.send_instance.sched_out
        assert(flag)
        self.assert_equal(msg_recv, self.test_msg)

    def test_call_alias(self):
        r"""Test RPC call aliases."""
        # self.send_instance.sched_task(0.0, self.send_instance.rpcSend,
        #                               args=[self.test_msg], store_output=True)
        self.recv_instance.sched_task(self.sleeptime, self.recv_instance.rpcRecv,
                                      kwargs=dict(timeout=self.timeout),
                                      store_output=True)
        flag = self.send_instance.rpcSend(self.test_msg)
        assert(flag)
        T = self.recv_instance.start_timeout()
        while (not T.is_out) and (self.recv_instance.sched_out is None):  # pragma: debug
            self.recv_instance.sleep()
        self.recv_instance.stop_timeout()
        flag, msg_recv = self.recv_instance.sched_out
        # flag, msg_recv = self.recv_instance.rpcRecv(timeout=self.timeout)
        assert(flag)
        self.assert_equal(msg_recv, self.test_msg)
        flag = self.recv_instance.rpcSend(msg_recv)
        assert(flag)
        flag, msg_recv = self.send_instance.rpcRecv(timeout=self.timeout)
        assert(flag)
        self.assert_equal(msg_recv, self.test_msg)

    def test_call_nolimit(self):
        r"""Test RPC nolimit call."""
        self.send_instance.sched_task(0.0, self.send_instance.call_nolimit,
                                      args=[self.msg_long], store_output=True)
        flag, msg_recv = self.recv_instance.recv_nolimit(timeout=self.timeout)
        assert(flag)
        self.assert_equal(msg_recv, self.msg_long)
        flag = self.recv_instance.send_nolimit(msg_recv)
        assert(flag)
        T = self.recv_instance.start_timeout()
        while (not T.is_out) and (self.send_instance.sched_out is None):  # pragma: debug
            self.recv_instance.sleep()
        self.recv_instance.stop_timeout()
        flag, msg_recv = self.send_instance.sched_out
        assert(flag)
        self.assert_equal(msg_recv, self.msg_long)

    def test_close_in_thread(self):
        r"""Test close of comm in thread."""
        self.send_instance.close_in_thread()
        self.recv_instance.close_in_thread()

    # # This dosn't work for comms that are uni-directional
    # def test_purge_recv(self):
    #     r"""Test purging messages from the client comm."""
    #     # Purge send while open
    #     if self.comm != 'CommBase':
    #         flag = self.send_instance.send(self.test_msg)
    #         assert(flag)
    #         T = self.recv_instance.start_timeout()
    #         while (not T.is_out) and (self.recv_instance.n_msg == 0):  # pragma: debug
    #             self.recv_instance.sleep()
    #         self.recv_instance.stop_timeout()
    #         self.assert_equal(self.recv_instance.n_msg, 1)
    #     self.send_instance.purge()
    #     self.assert_equal(self.send_instance.n_msg, 0)
    #     self.assert_equal(self.recv_instance.n_msg, 0)
    #     # Purge send while closed
    #     self.send_instance.close()
    #     self.send_instance.purge()



from __future__ import print_function
import sys
import numpy as np
from yggdrasil.interface.YggInterface import YggRpcClient


def fibClient(args):
    
    iterations = int(args[0])
    print('Hello from Python rpcFibCliPar: iterations = %d' % iterations)

    # Create RPC connection with server
    # RPC client-side connection will be $(server_name)_$(client_name)
    rpc = YggRpcClient("rpcFibSrv_rpcFibCliPar", "%d", "%d %d")

    # Send all of the requests to the server
    for i in range(1, iterations + 1):
        print('rpcFibCliPar(P): fib(->%-2d) ::: ' % i)
        ret = rpc.rpcSend(np.int32(i))
        if not ret:
            raise RuntimeError('rpcFibCliPar(P): SEND FAILED')

    # Receive responses for all requests that were sent
    for i in range(1, iterations + 1):
        ret, fib = rpc.rpcRecv()
        if not ret:
            raise RuntimeError('rpcFibCliPar(P): RECV FAILED')
        print('rpcFibCliPar(P): fib(%2d<-) = %-2d<-' % tuple(fib))

    print('Goodbye from Python rpcFibCliPar')

    
if __name__ == '__main__':
    fibClient(sys.argv[1:])



from yggdrasil.metaschema.properties.MetaschemaProperty import MetaschemaProperty


class ClassMetaschemaProperty(MetaschemaProperty):
    r"""Property class for 'class' property."""

    name = 'class'
    schema = {'description': ('One or more classes that the object should be '
                              'an instance of.'),
              'anyOf': [{'type': 'class'},
                        {'type': 'array',
                         'items': {'type': 'class'},
                         'minItems': 1}]}

    @classmethod
    def encode(cls, instance, typedef=None):
        r"""Encoder for the 'class' property."""
        return type(instance)

    @classmethod
    def _validate(cls, validator, value, instance, schema):
        r"""Validator for 'class' property."""
        # TODO: Normalization can be removed if metadata is normalized
        from yggdrasil.metaschema import validate_instance
        value = validate_instance(value, cls.schema, normalize=True)
        if not isinstance(instance, value):
            yield "Instance %s is not of type(s) %s" % (instance, value)
        
    @classmethod
    def compare(cls, prop1, prop2, root1=None, root2=None):
        r"""Method to determine compatiblity of one property value with another.
        This method is not necessarily symmetric in that the second value may
        not be compatible with the first even if the first is compatible with
        the second.

        Args:
            prop1 (object): Property value to compare against prop2.
            prop2 (object): Property value to compare against.
            
        Yields:
            str: Comparision failure messages.

        """
        # TODO: Normalization can be removed if metadata is normalized
        from yggdrasil.metaschema import validate_instance
        prop1 = validate_instance(prop1, cls.schema, normalize=True)
        prop2 = validate_instance(prop2, cls.schema, normalize=True)
        if not isinstance(prop1, (list, tuple)):
            prop1 = (prop1, )
        if not isinstance(prop2, (list, tuple)):
            prop2 = (prop2, )
        if not (set(prop1) & set(prop2)):
            yield 'No overlap between %s and %s' % (prop1, prop2)



import os
import copy
import unittest
from yggdrasil.tests import assert_raises, scripts
from yggdrasil.drivers.ModelDriver import ModelDriver, remove_product
from yggdrasil.drivers.CompiledModelDriver import CompiledModelDriver
from yggdrasil.drivers.InterpretedModelDriver import InterpretedModelDriver
import yggdrasil.drivers.tests.test_Driver as parent


def test_remove_product():
    r"""Test remove_product."""
    test_file = os.path.join(os.path.dirname(__file__), 'remove_product_test.py')
    with open(test_file, 'w') as fd:
        fd.write('print(\'hello\')')
    try:
        assert_raises(RuntimeError, remove_product, test_file,
                      check_for_source=True)
        assert_raises(RuntimeError, remove_product, os.path.dirname(test_file),
                      check_for_source=True)
    finally:
        os.remove(test_file)


def test_ModelDriver_implementation():
    r"""Test that NotImplementedError raised for base class."""
    assert_raises(NotImplementedError, ModelDriver.language_executable)
    assert_raises(NotImplementedError, ModelDriver.executable_command, None)
    assert_raises(NotImplementedError, ModelDriver.is_library_installed, None)
    assert_raises(NotImplementedError, CompiledModelDriver.get_tool, 'compiler')
    assert_raises(NotImplementedError, InterpretedModelDriver.get_interpreter)

    
class TestModelParam(parent.TestParam):
    r"""Test parameters for basic ModelDriver class."""

    driver = 'ModelDriver'
    
    def __init__(self, *args, **kwargs):
        super(TestModelParam, self).__init__(*args, **kwargs)
        self.attr_list += ['args', 'process', 'queue', 'queue_thread',
                           'is_server', 'client_of',
                           'event_process_kill_called',
                           'event_process_kill_complete',
                           'with_strace', 'strace_flags',
                           'with_valgrind', 'valgrind_flags',
                           'model_index', 'model_file', 'model_args',
                           'products', 'overwrite']
        self.src = None
        if self.import_cls.language is not None:
            self.src = scripts[self.import_cls.language]
            if not isinstance(self.src, list):
                self.src = [self.src]
        if self.src is not None:
            self.args = copy.deepcopy(self.src)

    def tests_on_not_installed(self):
        r"""Tests for when the driver is not installed."""
        if self.import_cls.is_installed():
            raise unittest.SkipTest("'%s' installed."
                                    % self.import_cls.language)

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        cls._flag_tests_on_not_installed = False
        super(TestModelParam, cls).setUpClass(*args, **kwargs)

    def setup(self, *args, **kwargs):
        if self.import_cls.language is None:
            raise unittest.SkipTest("Driver dosn't have language.")
        if not self.import_cls.is_installed():
            if not self.__class__._flag_tests_on_not_installed:
                if not self.skip_init:
                    self.assert_raises(RuntimeError,
                                       super(TestModelParam, self).setup,
                                       *args, **kwargs)
                self.tests_on_not_installed()
                self.__class__._flag_tests_on_not_installed = True
            raise unittest.SkipTest("'%s' not installed."
                                    % self.import_cls.language)
        super(TestModelParam, self).setup(*args, **kwargs)


class TestModelDriverNoInit(TestModelParam, parent.TestDriverNoInit):
    r"""Test runner for ModelDriver class without creating an instance."""

    def tests_on_not_installed(self):
        r"""Tests for when the driver is not installed."""
        super(TestModelDriverNoInit, self).tests_on_not_installed()
        self.test_comm_installed()
        self.test_write_if_block()
        self.test_write_for_loop()
        self.test_write_while_loop()
        self.test_write_try_except()

    def test_is_installed(self):
        r"""Assert that the tested model driver is installed."""
        assert(self.import_cls.is_installed())

    def test_comm_installed(self):
        r"""Tests for getting installed comm while skipping config."""
        self.assert_equal(self.import_cls.is_comm_installed(),
                          self.import_cls.is_comm_installed(skip_config=True))
        self.assert_equal(self.import_cls.is_comm_installed(commtype='invalid',
                                                            skip_config=True),
                          False)
        
    def test_language_version(self):
        r"""Test language version."""
        assert(self.import_cls.language_version())

    # Tests for code generation
    def run_generated_code(self, lines):
        r"""Write and run generated code."""
        if not self.import_cls.is_installed():
            return
        # Write code to a file
        self.import_cls.run_code(lines)
            
    def test_write_executable(self):
        r"""Test writing an executable."""
        if self.import_cls.function_param is None:
            self.assert_raises(NotImplementedError,
                               self.import_cls.write_executable,
                               None)
        else:
            lines1 = self.import_cls.write_executable('dummy',
                                                      prefix='dummy',
                                                      suffix='dummy')
            lines2 = self.import_cls.write_executable(lines1)
            self.assert_equal(lines1, lines2)
            # Don't run this because it is invalid

    def test_error_code(self):
        r"""Test that error is raised when code generates one."""
        if (((not self.import_cls.is_installed())
             or (self.import_cls.function_param is None))):
            return
        error_msg = 'Test error'
        lines = [self.import_cls.function_param['error'].format(error_msg=error_msg)]
        assert_raises(RuntimeError, self.import_cls.run_code, lines)

    def test_write_if_block(self):
        r"""Test writing an if block."""
        if self.import_cls.function_param is None:
            self.assert_raises(NotImplementedError, self.import_cls.write_if_block,
                               None, None)
        else:
            lines = []
            if 'declare' in self.import_cls.function_param:
                lines.append(self.import_cls.function_param['declare'].format(
                    type='int', name='x'))
            cond = self.import_cls.function_param['true']
            block_contents = self.import_cls.function_param['assign'].format(
                name='x', value='1')
            lines += self.import_cls.write_if_block(cond, block_contents)
            self.run_generated_code(lines)

    def test_write_for_loop(self):
        r"""Test writing a for loop."""
        if self.import_cls.function_param is None:
            self.assert_raises(NotImplementedError, self.import_cls.write_for_loop,
                               None, None, None, None)
        else:
            lines = []
            if 'declare' in self.import_cls.function_param:
                lines.append(self.import_cls.function_param['declare'].format(
                    type='int', name='i'))
                lines.append(self.import_cls.function_param['declare'].format(
                    type='int', name='x'))
            loop_contents = self.import_cls.function_param['assign'].format(
                name='x', value='i')
            lines += self.import_cls.write_for_loop('i', 0, 1, loop_contents)
            self.run_generated_code(lines)

    def test_write_while_loop(self):
        r"""Test writing a while loop."""
        if self.import_cls.function_param is None:
            self.assert_raises(NotImplementedError, self.import_cls.write_while_loop,
                               None, None)
        else:
            lines = []
            cond = self.import_cls.function_param['true']
            loop_contents = self.import_cls.function_param.get('break', 'break')
            lines += self.import_cls.write_while_loop(cond, loop_contents)
            self.run_generated_code(lines)

    def test_write_try_except(self, **kwargs):
        r"""Test writing a try/except block."""
        if self.import_cls.function_param is None:
            self.assert_raises(NotImplementedError, self.import_cls.write_try_except,
                               None, None)
        else:
            lines = []
            try_contents = self.import_cls.function_param['error'].format(
                error_msg='Dummy error')
            except_contents = self.import_cls.function_param['print'].format(
                message='Dummy message')
            lines += self.import_cls.write_try_except(try_contents,
                                                      except_contents, **kwargs)
            self.run_generated_code(lines)
    

class TestModelDriverNoStart(TestModelParam, parent.TestDriverNoStart):
    r"""Test runner for basic ModelDriver class."""
    pass


class TestModelDriver(TestModelParam, parent.TestDriver):
    r"""Test runner for basic ModelDriver class."""
    pass



from collections import OrderedDict
from yggdrasil.metaschema import encoder
from yggdrasil.backwards import StringIO
from yggdrasil.tests import assert_raises


class TestClass(object):  # pragma: no cover
    pass


def test_JSONEncoder():
    r"""Test JSONEncoder error."""
    x = TestClass()
    assert_raises(TypeError, encoder.encode_json, x)


def test_encode_yaml():
    r"""Test encode_yaml with dict representer and file."""
    x = OrderedDict([('a', 1), ('b', 2)])
    fd = StringIO()
    encoder.encode_yaml(x, fd=fd, sorted_dict_type=OrderedDict)
    fd.close()



from yggdrasil.communication.AsciiTableComm import AsciiTableComm


class PandasFileComm(AsciiTableComm):
    r"""Class for handling I/O from/to a pandas csv file on disk.

    Args:
        name (str): The environment variable where communication address is
            stored.
        delimiter (str, optional): String that should be used to separate
            columns. Defaults to '\t'.
        **kwargs: Additional keywords arguments are passed to parent class.

    """

    _filetype = 'pandas'
    _schema_subtype_description = ('The file is a Pandas frame output as a table.')
    _default_serializer = 'pandas'



# Normalizer adapated from the jsonschema validator
import pprint
import copy
import contextlib
import jsonschema
import logging
from collections import OrderedDict
from yggdrasil.metaschema.datatypes import get_type_class, _jsonschema_ver_maj


logger = logging.getLogger(__name__)


class UndefinedProperty(object):
    r"""Class to be used as a flag for undefined properties."""
    pass


class UninitializedNormalized(object):
    r"""Class to be used as a flag for uninitialized normalized value."""
    pass


def create(*args, **kwargs):
    r"""Dynamically create a validation/normalization class that subclasses
    the jsonschema validation class.

    Args:
        normalizers (dict, optional): Keys are tuples representing paths that
            exist within the schema at which the normalization functions stored
            in lists as their value counterparts should be executed. Defaults to
            empty dictionary.
        no_defaults (bool, optional): If True, defaults will not be set during
            normalization. Defaults to False.
        required_defaults (bool, optional): If True, defaults will be set for
            required properties, even if no_defaults is True. Defaults to False.
        *args: Additional arguments are passed to jsonschema.validators.create.
        **kwargs: Additional keyword arguments are passed to
            jsonschema.validators.create.
        
    """
    normalizers = kwargs.pop('normalizers', ())
    no_defaults = kwargs.pop('no_defaults', ())
    required_defaults = kwargs.pop('required_defaults', ())
    validator_class = jsonschema.validators.create(*args, **kwargs)

    class Normalizer(validator_class):
        r"""Class that can be used to normalize (or validate) objects against
        JSON schemas.

        Args:
            *args: Additional arguments are passed to the base validator class.
            **kwargs: Additional keyword arguments are passed to the base
                validator class.

        Attributes:
            NORMALIZERS (dict): Keys are tuples representing paths that exist
                within the schema at which the normalization functions stored in
                lists as their value counterparts should be executed.
            NO_DEFAULTS (bool): If True, defaults will not be set during
                normalization.
            REQUIRED_DEFAULTS (bool): If True, defaults will be set for required
                properties, even if NO_DEFAULTS is True.

        """
        NORMALIZER_VALIDATORS = OrderedDict([('default', None),
                                             ('type', None)])
        NORMALIZERS = dict(normalizers)
        NO_DEFAULTS = no_defaults
        REQUIRED_DEFAULTS = required_defaults
        VERBOSE = False

        def __init__(self, *args, **kwargs):
            super(Normalizer, self).__init__(*args, **kwargs)
            self._normalized = UninitializedNormalized()
            self._normalizing = False
            self._old_settings = {}
            self._path_stack = []
            self._schema_path_stack = []
            self._normalized_stack = []

        @classmethod
        def normalize_schema(cls, schema, **kwargs):
            r"""Normalize a schema against the metaschema.

            Args:
                schema (dict): Schema that should be normalized.
                **kwargs: Additional keyword arguments are passed to the
                    normalize method.

            Returns:
                dict: Normalized schema.

            """
            return cls(cls.META_SCHEMA).normalize(schema, **kwargs)

        @property
        def current_path(self):
            r"""tuple: Current path from the top of the instance to the current
            instance being validated/normalized."""
            return tuple(self._path_stack)

        @property
        def current_schema_path(self):
            r"""tuple: Current path from the top of the schema to the current
            schema being used for validation/normalization."""
            return tuple(self._schema_path_stack)

        @contextlib.contextmanager
        def normalizing(self, **kwargs):
            r"""Context for normalization that records normalizers before
            context is initialized so that they can be restored once the context
            exist.

            Args:
                **kwargs: Keyword arguments are treated as attributes that
                    should be added to the class in the context. If the class
                    already has an attribute of the same name, it is stored
                    for restoration after the context exits.

            Yields:
                ValidationError: Errors encountered during validation.

            """
            kwargs.update(iter_errors=self.iter_errors_normalize,
                          descend=self.descend_normalize,
                          _normalizing=True)
            for k, v in kwargs.items():
                if k == 'normalizers':
                    if self.NORMALIZERS:  # pragma: debug
                        raise Exception("Uncomment lines below to allow "
                                        + "addition of default normalizers.")
                    # for ik, iv in self.NORMALIZERS.items():
                    #     v.setdefault(ik, iv)
                elif k == 'validators':
                    for ik, iv in self.VALIDATORS.items():
                        v.setdefault(ik, iv)
                elif k == 'normalizer_validators':
                    for ik, iv in self.NORMALIZER_VALIDATORS.items():
                        v.setdefault(ik, iv)
                if hasattr(self, k.upper()):
                    ksub = k.upper()
                else:
                    ksub = k
                self._old_settings[ksub] = getattr(self, ksub, None)
                setattr(self, ksub, v)
            # Separate out validators that need to be run in a specific order
            # during normalization
            _migrated_validators = []
            for k in self.NORMALIZER_VALIDATORS.keys():
                if self.NORMALIZER_VALIDATORS[k] is None:
                    _migrated_validators.append(k)
                    self.NORMALIZER_VALIDATORS[k] = self.VALIDATORS.get(k, None)
                    self.VALIDATORS[k] = None
            # Perform context and then cleanup
            try:
                yield
            finally:
                # Restore validators with special order
                for k in _migrated_validators:
                    self.VALIDATORS[k] = self.NORMALIZER_VALIDATORS[k]
                    self.NORMALIZER_VALIDATORS[k] = None
                # Restore old attributes
                for k, v in self._old_settings.items():
                    if v is None:
                        delattr(self, k)
                    else:
                        setattr(self, k, v)
                self._old_settings = {}

        def iter_errors_normalize(self, instance, _schema=None):
            r"""Iterate through all of the errors encountered during validation
            of an instance at the current level or lower against properties in a
            schema.

            Args:
                instance (object): Instance that will be validated.
                _schema (dict, optional): Schema that the instance will be
                    validated against. Defaults to the schema used to initialize
                    the class.

            Yields:
                ValidationError: Errors encountered during validation of the
                    instance.

            """
            if _schema is None:
                _schema = self.schema

            if isinstance(self._normalized, UninitializedNormalized):
                self._normalized = copy.deepcopy(instance)

            if isinstance(_schema, dict) and (u"$ref" not in _schema):
                # Path based normalization
                try:
                    # logger.info("schema_path=%s, type=%s, instance=%s, schema=%s"
                    #             % (self.current_schema_path,
                    #                type(self._normalized), self._normalized,
                    #                _schema))
                    if self.current_schema_path in self.NORMALIZERS:
                        normalizers = self.NORMALIZERS[self.current_schema_path]
                        for n in normalizers:
                            self._normalized = n(self, None, self._normalized, _schema)
                except BaseException as e:
                    error = jsonschema.ValidationError(str(e))
                    # set details if not already set by the called fn
                    error._set(
                        validator=n,
                        validator_value=None,
                        instance=self._normalized,
                        schema=_schema)
                    # if self.VERBOSE:  # pragma: debug
                    #     logger.info('Error in normalization: %s' % e)
                    yield error

                # Do defaults for required fields
                if (((((not self.NO_DEFAULTS) or self.REQUIRED_DEFAULTS)
                      and isinstance(_schema.get('required', None), list)
                      and isinstance(_schema.get('properties', None), dict)
                      and self.is_type(self._normalized, "object")))):
                    for k in _schema['required']:
                        if (((k not in _schema['properties'])
                             or (k in self._normalized))):
                            continue
                        default = _schema['properties'][k].get('default', None)
                        self._normalized[k] = copy.deepcopy(default)

                # Perform normalization for properties that will change the
                # outcome of validation
                for k, validator in self.NORMALIZER_VALIDATORS.items():
                    if (((k != 'default')
                         and isinstance(self._normalized, UndefinedProperty))):
                        return
                    if (validator is None) or (k not in _schema):
                        continue
                    v = _schema[k]
                    errors = validator(self, v, self._normalized, _schema) or ()
                    for error in errors:
                        # set details if not already set by the called fn
                        error._set(
                            validator=k,
                            validator_value=v,
                            instance=self._normalized,
                            schema=_schema,
                        )
                        if k != u"$ref":
                            error.schema_path.appendleft(k)
                        # if self.VERBOSE:  # pragma: debug
                        #     logger.info('Error in early %s validation: %s'
                        #                 % (k, error))
                        yield error

            for e in self._old_settings['iter_errors'](self._normalized,
                                                       _schema=_schema):
                # if self.VERBOSE:  # pragma: debug
                #     logger.info('Error in base iter_errors: %s' % e)
                yield e

        def descend_normalize(self, instance, schema, path=None, schema_path=None):
            r"""Descend along a path in the schema/instance, recording
            information about the normalization state so that it can be replaced
            with the original value if there is a validation error along the
            descent path.

            Args:
                instance (object): Current instance being validated against the
                    schema.
                schema (dict): Current schema that the instance is being
                    validated against.
                path (str, int, optional): Path that resulted in the current
                    instance. Defaults to None.
                schema_path (str, int, optional): Path that resulted in the
                    current schema. Defaults to None.

            Yields:
                ValidationError: Errors raised during validation of the instance.

            """
            old_normalized = self._normalized
            if path is not None:
                # self._normalized_stack.append(self._normalized)
                self._normalized = UninitializedNormalized()
            else:
                # self._normalized_stack.append(self._normalized)
                self._normalized = copy.deepcopy(self._normalized)
            if path is not None:
                self._path_stack.append(path)
            if schema_path is not None:
                self._schema_path_stack.append(schema_path)
            failed = False
            try:
                for error in self._old_settings['descend'](instance, schema,
                                                           path=path,
                                                           schema_path=schema_path):
                    failed = True
                    # if self.VERBOSE:
                    #     logger.info("Error in descent (path=%s, schema_path=%s): %s"
                    #                 % (path, schema_path, error))
                    yield error
            finally:
                # old_normalized = self._normalized_stack.pop()
                if not (failed or isinstance(self._normalized, UndefinedProperty)):
                    if path is not None:
                        old_normalized[path] = self._normalized
                    else:
                        old_normalized = self._normalized
                self._normalized = old_normalized
                if path is not None:
                    self._path_stack.pop()
                if schema_path is not None:
                    self._schema_path_stack.pop()

        def validate(self, instance, _schema=None, normalize=False, **kwargs):
            r"""Validate an instance against a schema.

            Args:
                instance (object): Object to be validated.
                _schema (dict, optional): Schema by which the instance should be
                    validated. Defaults to None and will be set to the schema
                    used to create the class.
                normalize (bool, optional): If True, the instance will also be
                    normalized as it is validated. Defaults to False.
                **kwargs: Additional keyword arguments are passed to the
                    'normalizing' context if normalize is True, otherwise they
                    are ignored.

            Returns:
                object: Normalized instance if normalize == True.

            """
            if normalize:
                with self.normalizing(**kwargs):
                    super(Normalizer, self).validate(instance, _schema=_schema)
                out = self._normalized
                return out
            else:
                super(Normalizer, self).validate(instance, _schema=_schema)

        def normalize(self, instance, _schema=None, show_errors=False, **kwargs):
            r"""Normalize an instance during validation, allowing for aliases,
            defaults, or simple type conversions.

            Args:
                instance (object): Object to be normalized and validated.
                _schema (dict, optional): Schema by which the instance should be
                    normalized and validated. Defaults to None and will be set
                    to the schema used to create the class.
                show_errors (bool, optional): If True, any errors during the
                    normalization are displayed. Defaults to False.
                **kwargs: Additional keyword arguments are passed to the
                    'normalizing' context.

            Returns:
                object: Normalized instance.

            """
            if show_errors:
                self.VERBOSE = True
            with self.normalizing(**kwargs):
                errors = list(self.iter_errors(instance, _schema=_schema))
                if show_errors:  # pragma: debug
                    for e in errors[::-1]:
                        if e:
                            logger.info(80 * '-')
                            logger.info(e)
                    if errors:
                        logger.info(80 * '-')
                        logger.info('Normalized:\n'
                                    + pprint.pformat(self._normalized))
            if errors:
                return instance
            else:
                return self._normalized

    def iter_errors_js2(self, instance, _schema=None):
        r"""Wrapper that can be added to classes for jsonschema < 3.0 to allow
        for boolean schema."""
        if _schema is None:
            _schema = self.schema
        if _schema is True:
            pass
        elif _schema is False:
            yield jsonschema.exceptions.ValidationError(
                "False schema does not allow %r" % (instance,),
                validator=None,
                validator_value=None,
                instance=instance,
                schema=_schema)
        else:
            for e in super(Normalizer, self).iter_errors(instance, _schema=_schema):
                yield e

    def is_type_js2(self, instance, types):
        r"""Jsonschema < 3.0 wrapper for method that determines if an object is
        an example of the given type.

        Args:
            instance (object): Object to test against to the type.
            type (str, list): Name of single type or a list of types that
                instance should be tested against.

        Returns:
            bool: True if the instance is of the specified type(s). False
                otherwise.

        """
        out = super(Normalizer, self).is_type(instance, types)
        if out:
            out = get_type_class(types).validate(instance)
        return out

    if _jsonschema_ver_maj < 3:
        Normalizer.iter_errors = iter_errors_js2
        Normalizer.is_type = is_type_js2
    return Normalizer



import sys
import random
from yggdrasil.interface.YggInterface import YggRpcClient, YGG_MSG_BUF


def rand_str(length):
    charset = ("0123456789"
               + "abcdefghijklmnopqrstuvwxyz"
               + "ABCDEFGHIJKLMNOPQRSTUVWXYZ")
    out = ''
    while (len(out) < (length - 1)):
        index = int(random.random() * len(charset))
        out += charset[index]
    if sys.version_info[0] == 3:
        out = out.encode("utf-8")
    return out


msg_size = YGG_MSG_BUF

print("maxMsgCli(P): Hello message size %d." % msg_size)

# Create a max message, send/recv and verify
rpc = YggRpcClient("maxMsgSrv_maxMsgCli", "%s", "%s")

# Create a max message
output = rand_str(msg_size)

# Call RPC server
ret, input = rpc.rpcCall(output)
if not ret:
    raise RuntimeError("maxMsgCli(P): RPC ERROR")

# Check to see if response matches
if (input[0] != output):
    raise RuntimeError("maxMsgCli(P): ERROR: input/output do not match")
else:
    print("maxMsgCli(P): CONFIRM")

# All done, say goodbye
print("maxMsgCli(P): Goodbye!")



import unittest
from yggdrasil.communication.RMQComm import RMQComm
import yggdrasil.drivers.tests.test_ClientDriver as parent


_rmq_installed = RMQComm.is_installed(language='python')


@unittest.skipIf(not _rmq_installed, "RMQ Server not running")
class TestRMQAsyncClientParam(parent.TestClientParam):
    r"""Test parameters for RMQAsyncClientDriver class."""

    def __init__(self, *args, **kwargs):
        super(TestRMQAsyncClientParam, self).__init__(*args, **kwargs)
        self.driver = 'RMQAsyncClientDriver'
        self.server_comm = 'RMQAsyncComm'
        self.ocomm_name = self.server_comm
        self.timeout = 10.0
        self.route_timeout = 60.0

    
@unittest.skipIf(not _rmq_installed, "RMQ Server not running")
class TestRMQAsyncClientDriverNoStart(TestRMQAsyncClientParam,
                                      parent.TestClientDriverNoStart):
    r"""Test class for RMQAsyncClientDriver class without start."""
    pass


@unittest.skipIf(not _rmq_installed, "RMQ Server not running")
class TestRMQAsyncClientDriver(TestRMQAsyncClientParam, parent.TestClientDriver):
    r"""Test class for RMQAsyncClientDriver class."""
    pass



r"""Tests for platform compatiblity."""
from yggdrasil.tests import assert_equal
from yggdrasil import platform


def test_os():
    r"""Test that only one OS selected."""
    res = platform._is_mac + platform._is_linux + platform._is_win
    assert_equal(res, 1)



import unittest
from yggdrasil import platform
from yggdrasil.tests import assert_raises
from yggdrasil.drivers.ExecutableModelDriver import ExecutableModelDriver
import yggdrasil.drivers.tests.test_ModelDriver as parent


def test_execution_error():
    r"""Test errors raised during execution of invalid command."""
    assert_raises(RuntimeError, ExecutableModelDriver.run_executable,
                  ['invalid'])


def test_error_valgrind_strace():
    r"""Test error if both valgrind and strace set."""
    assert_raises(RuntimeError, ExecutableModelDriver, 'test', 'test',
                  with_strace=True, with_valgrind=True)


@unittest.skipIf(not platform._is_win, "Platform is not windows")
def test_error_valgrind_strace_windows():  # pragma: windows
    r"""Test error if strace or valgrind called on windows."""
    assert_raises(RuntimeError, ExecutableModelDriver, 'test', 'test',
                  with_strace=True)
    assert_raises(RuntimeError, ExecutableModelDriver, 'test', 'test',
                  with_valgrind=True)

    
class TestExecutableModelParam(parent.TestModelParam):
    r"""Test parameters for ExecutableModelDriver class."""

    driver = 'ExecutableModelDriver'

            
class TestExecutableModelDriverNoInit(TestExecutableModelParam,
                                      parent.TestModelDriverNoInit):
    r"""Test runner for ExecutableModelDriver class without init."""
    pass


class TestExecutableModelDriverNoStart(TestExecutableModelParam,
                                       parent.TestModelDriverNoStart):
    r"""Test runner for ExecutableModelDriver class without start."""
    pass


class TestExecutableModelDriver(TestExecutableModelParam,
                                parent.TestModelDriver):
    r"""Test runner for ExecutableModelDriver class."""
    pass
                           

@unittest.skipIf(platform._is_win, "Platform is windows")
class TestExecutableModelDriver_valgrind(TestExecutableModelDriver):
    r"""Test with valgrind."""

    @property
    def inst_kwargs(self):
        r"""dict: Keyword arguments for creating a class instance."""
        out = super(TestExecutableModelDriver_valgrind, self).inst_kwargs
        out['with_valgrind'] = True
        return out


@unittest.skipIf(platform._is_win, "Platform is windows")
class TestExecutableModelDriver_strace(TestExecutableModelDriver):
    r"""Test with strace."""

    @property
    def inst_kwargs(self):
        r"""dict: Keyword arguments for creating a class instance."""
        out = super(TestExecutableModelDriver_strace, self).inst_kwargs
        out['with_strace'] = True
        return out



import os
from yggdrasil.tests import extra_example
from yggdrasil.examples.tests import TestExample


@extra_example
class TestExampleSaM(TestExample):
    r"""Test the SaM example."""

    example_name = 'SaM'

    @property
    def results(self):
        r"""list: Results that should be found in the output files."""
        # 1 + 2*n_languages
        if self.language == 'all':  # pragma: matlab
            s = 9
        elif self.language == 'all_nomatlab':  # pragma: no matlab
            s = 7
        else:
            s = 3
        return ['%d' % s]

    @property
    def output_files(self):
        r"""list: Output files for the run."""
        return [os.path.join(self.tempdir, 'SaM_output.txt')]



import os


_lang_dir = os.path.dirname(__file__)
_top_dir = os.path.normpath(os.path.join(_lang_dir, '../'))


def get_language_dir(lang):
    r"""Return the directory containing the interface for the requested language.

    Args:
        lang (str): The name of the language to return a directory for.

    Returns:
        str: The full path to the interface directory for the language.

    """
    mod_list = [lang, lang.lower(), lang.upper(), lang.title()]
    if lang.lower() in ['cmake', 'lpy']:
        mod_list.insert(0, lang[:2].upper() + lang[2:].lower())
    elif lang.lower() == 'cpp':
        mod_list = ['c++', 'C++'] + mod_list
    elif lang.lower() == 'c++':
        mod_list += ['cpp', 'CPP']
    for ilang in mod_list:
        idir = os.path.join(_lang_dir, ilang)
        if os.path.isdir(idir):
            return idir
    raise ValueError("Could not determine directory for the language: '%s'" % lang)



import numpy as np
from yggdrasil import units, backwards, platform
from yggdrasil.metaschema.datatypes import MetaschemaTypeError
from yggdrasil.metaschema.properties.MetaschemaProperty import MetaschemaProperty
from collections import OrderedDict


_valid_numpy_types = ['int', 'uint', 'float', 'complex']
_valid_types = OrderedDict([(k, k) for k in _valid_numpy_types])
_flexible_types = ['string', 'bytes', 'unicode']
_python_scalars = OrderedDict([('float', [float]),
                               ('int', [int]), ('uint', []),
                               ('complex', [complex])])
if backwards.PY2:  # pragma: Python 2
    from __builtin__ import unicode
    _valid_numpy_types += ['string', 'unicode']
    _valid_types['bytes'] = 'string'
    _valid_types['unicode'] = 'unicode'
    _python_scalars.update([('bytes', [str]), ('unicode', [unicode])])
else:  # pragma: Python 3
    _valid_numpy_types += ['bytes', 'unicode', 'str']
    _valid_types['bytes'] = 'bytes'
    _valid_types['unicode'] = 'str'
    _python_scalars.update([('bytes', [bytes]), ('unicode', [str])])
for t, t_np in _valid_types.items():
    prec_list = []
    if t in ['float']:
        prec_list = [16, 32, 64]
        if not platform._is_win:
            prec_list.append(128)  # Not available on windows
    if t in ['int', 'uint']:
        prec_list = [8, 16, 32, 64]
    elif t in ['complex']:
        prec_list = [64, 128]
        if not platform._is_win:
            prec_list.append(256)  # Not available on windows
    if hasattr(np, t_np):
        _python_scalars[t].append(getattr(np, t_np))
    _python_scalars[t].append(np.dtype(t_np).type)
    for p in prec_list:
        _python_scalars[t].append(np.dtype(t_np + str(p)).type)
# For some reason windows fails to check types on ints in some cases
_python_scalars['int'].append(np.signedinteger)
_python_scalars['uint'].append(np.unsignedinteger)
_all_python_scalars = [units._unit_quantity]
for k in _python_scalars.keys():
    _python_scalars[k].append(units._unit_quantity)
    _all_python_scalars += list(_python_scalars[k])
    _python_scalars[k] = tuple(_python_scalars[k])
_all_python_arrays = tuple(set([np.ndarray, units._unit_array]))
_all_python_scalars = tuple(set(_all_python_scalars))


def data2dtype(data):
    r"""Get numpy data type for an object.

    Args:
        data (object): Python object.

    Returns:
        np.dtype: Numpy data type.

    """
    data_nounits = units.get_data(data)
    if isinstance(data_nounits, np.ndarray):
        dtype = data_nounits.dtype
    elif isinstance(data_nounits, (list, dict, tuple)):
        raise MetaschemaTypeError
    elif isinstance(data_nounits, np.dtype(_valid_types['bytes']).type):
        dtype = np.array(data_nounits).dtype
    else:
        dtype = np.array([data_nounits]).dtype
    return dtype


def definition2dtype(props):
    r"""Get numpy data type for a type definition.

    Args:
        props (dict): Type definition properties.
        
    Returns:
        np.dtype: Numpy data type.

    """
    typename = props.get('subtype', None)
    if typename is None:
        typename = props.get('type', None)
        if typename is None:
            raise KeyError('Could not find type in dictionary')
    if ('precision' not in props):
        if typename in _flexible_types:
            out = np.dtype((_valid_types[typename]))
        else:
            raise RuntimeError("Precision required for type: '%s'" % typename)
    elif typename == 'unicode':
        out = np.dtype((_valid_types[typename], int(props['precision'] // 32)))
    elif typename in _flexible_types:
        out = np.dtype((_valid_types[typename], int(props['precision'] // 8)))
    else:
        out = np.dtype('%s%d' % (_valid_types[typename], int(props['precision'])))
    return out


class SubtypeMetaschemaProperty(MetaschemaProperty):
    r"""Property class for 'subtype' property."""

    name = 'subtype'
    schema = {'description': 'The base type for each item.',
              'type': 'string',
              'enum': [k for k in sorted(_valid_types.keys())]}

    @classmethod
    def encode(cls, instance, typedef=None):
        r"""Encoder for the 'subtype' scalar property."""
        dtype = data2dtype(instance)
        out = None
        for k, v in _valid_types.items():
            if dtype.name.startswith(v):
                out = k
                break
        if out is None:
            raise MetaschemaTypeError('Cannot find subtype string for dtype %s'
                                      % dtype)
        return out

    @classmethod
    def normalize_in_schema(cls, schema):
        r"""Normalization for the 'subtype' scalar property in a schema."""
        if cls.name in schema:
            return schema
        if not units.is_null_unit(schema.get('units', '')):
            schema.setdefault(cls.name, 'float')
        return schema
    

class PrecisionMetaschemaProperty(MetaschemaProperty):
    r"""Property class for 'precision' property."""
    
    name = 'precision'
    schema = {'description': 'The size (in bits) of each item.',
              'type': 'number',
              'minimum': 1}

    @classmethod
    def encode(cls, instance, typedef=None):
        r"""Encoder for the 'precision' scalar property."""
        dtype = data2dtype(instance)
        out = dtype.itemsize * 8  # in bits
        return out

    @classmethod
    def compare(cls, prop1, prop2, root1=None, root2=None):
        r"""Comparison for the 'precision' scalar property."""
        if (prop1 > prop2):
            yield '%s is greater than %s' % (prop1, prop2)

    @classmethod
    def normalize_in_schema(cls, schema):
        r"""Normalization for the 'precision' scalar property in a schema."""
        if cls.name in schema:
            return schema
        subtype = schema.get('subtype', None)
        if subtype in ['float', 'int', 'uint']:
            schema.setdefault(cls.name, int(64))
        elif subtype in ['complex']:
            schema.setdefault(cls.name, int(128))
        return schema
            

class UnitsMetaschemaProperty(MetaschemaProperty):
    r"""Property class for 'units' property."""

    name = 'units'
    schema = {'description': 'Physical units.',
              'type': 'string'}

    @classmethod
    def encode(cls, instance, typedef=None):
        r"""Encoder for the 'units' scalar property."""
        out = units.get_units(instance)
        if (not out) and (typedef is not None):
            out = typedef
        return out

    @classmethod
    def compare(cls, prop1, prop2, root1=None, root2=None):
        r"""Comparision for the 'units' scalar property."""
        if not units.are_compatible(prop1, prop2):
            yield "Unit '%s' is not compatible with unit '%s'" % (prop1, prop2)



"""Testing things."""
import os
import shutil
import uuid
import difflib
import importlib
import contextlib
import warnings
import unittest
import numpy as np
import pandas as pd
import threading
import psutil
import copy
import pprint
from yggdrasil.config import ygg_cfg, cfg_logging
from yggdrasil import tools, backwards, platform, units
from yggdrasil.communication import cleanup_comms
from yggdrasil.components import import_component


# Test data
data_dir = os.path.join(os.path.dirname(__file__), 'data')
data_list = [
    ('txt', 'ascii_file.txt'),
    ('table', 'ascii_table.txt')]
data = {k: os.path.join(data_dir, v) for k, v in data_list}

# Test scripts
script_dir = os.path.join(os.path.dirname(__file__), 'scripts')
script_list = [
    ('c', ['gcc_model.c', 'hellofunc.c']),
    ('c++', ['gcc_model.cpp', 'hellofunc.c']),
    ('make', 'gcc_model'),
    ('cmake', 'gcc_model'),
    ('matlab', 'matlab_model.m'),
    ('matlab_error', 'matlab_error_model.m'),
    ('python', 'python_model.py'),
    ('error', 'error_model.py'),
    ('lpy', 'lpy_model.lpy')]
scripts = {}
for k, v in script_list:
    if isinstance(v, list):
        scripts[k] = [os.path.join(script_dir, iv) for iv in v]
    else:
        scripts[k] = os.path.join(script_dir, v)
# scripts = {k: os.path.join(script_dir, v) for k, v in script_list}
if platform._is_win:  # pragma: windows
    scripts['executable'] = ['timeout', '0']
else:
    scripts['executable'] = ['sleep', 0.1]
    
# Test yamls
yaml_dir = os.path.join(os.path.dirname(__file__), 'yamls')
yaml_list = [
    ('c', 'gcc_model.yml'),
    ('cpp', 'gpp_model.yml'),
    ('make', 'make_model.yml'),
    ('cmake', 'cmake_model.yml'),
    ('matlab', 'matlab_model.yml'),
    ('python', 'python_model.yml'),
    ('error', 'error_model.yml'),
    ('lpy', 'lpy_model.yml')]
yamls = {k: os.path.join(yaml_dir, v) for k, v in yaml_list}

# Makefile
if platform._is_win:  # pragma: windows
    makefile0 = os.path.join(script_dir, "Makefile_windows")
else:
    makefile0 = os.path.join(script_dir, "Makefile_linux")
shutil.copy(makefile0, os.path.join(script_dir, "Makefile"))


# Flag for enabling tests that take a long time or are for extra examples
enable_long_tests = tools.check_environ_bool("YGG_ENABLE_LONG_TESTS")
skip_extra_examples = tools.check_environ_bool("YGG_SKIP_EXTRA_EXAMPLES")


# Wrapped class to allow handling of arrays
class WrappedTestCase(unittest.TestCase):  # pragma: no cover
    def __init__(self, *args, **kwargs):
        super(WrappedTestCase, self).__init__(*args, **kwargs)
        self.addTypeEqualityFunc(units._unit_quantity, 'assertUnitsEqual')
        self.addTypeEqualityFunc(units._unit_array, 'assertUnitsEqual')
        self.addTypeEqualityFunc(np.ndarray, 'assertArrayEqual')
        self.addTypeEqualityFunc(pd.DataFrame, 'assertArrayEqual')

    def has_units(self, obj):
        if isinstance(obj, (list, tuple)):
            for x in obj:
                if self.has_units(x):
                    return True
        elif isinstance(obj, dict):
            for x in obj.values():
                if self.has_units(x):
                    return True
        else:
            return units.has_units(obj)
        return False

    def _getAssertEqualityFunc(self, first, second):
        # Allow comparison of tuple to list and units to anything
        if (type(first), type(second)) in [(list, tuple), (tuple, list)]:
            return self.assertSequenceEqual
        elif units.has_units(first) or units.has_units(second):
            return self.assertUnitsEqual
        return super(WrappedTestCase, self)._getAssertEqualityFunc(first, second)
        
    def assertEqual(self, first, second, msg=None, dont_nest=False):
        r"""Fail if the two objects are unequal as determined by the '=='
           operator."""
        if (not dont_nest):
            # Do nested evaluation for objects containing units
            if (self.has_units(first) or self.has_units(second)):
                self.assertEqualNested(first, second, msg=msg)
                return
        try:
            super(WrappedTestCase, self).assertEqual(first, second, msg=msg)
        except ValueError:
            if dont_nest:
                raise
            self.assertEqualNested(first, second, msg=msg)

    def assertEqualNested(self, first, second, msg=None):
        r"""Fail if the two objects are unequal as determined by descending
        recursively into the object if it is a list, tuple, or dictionary."""
        if isinstance(first, list):
            self.assertSequenceEqualNested(first, second, msg=msg, seq_type=list)
        elif isinstance(first, tuple):
            self.assertSequenceEqualNested(first, second, msg=msg, seq_type=tuple)
        elif isinstance(first, dict):
            self.assertDictEqualNested(first, second, msg=msg)
        else:
            self.assertEqual(first, second, msg=msg, dont_nest=True)
    
    def assertSequenceEqualNested(self, seq1, seq2, msg=None, seq_type=None):
        if seq_type is not None:
            seq_type_name = seq_type.__name__
            # Currently it makes more sense to allow equality between lists and
            # tuples
            # if not isinstance(seq1, seq_type):
            #     raise self.failureException(
            #         'First sequence is not a %s: %s'
            #         % (seq_type_name, unittest.util.safe_repr(seq1)))
            # if not isinstance(seq2, seq_type):
            #     raise self.failureException(
            #         'Second sequence is not a %s: %s'
            #         % (seq_type_name, unittest.util.safe_repr(seq2)))
        else:
            seq_type_name = "sequence"

        differing = None
        try:
            len1 = len(seq1)
        except (TypeError, NotImplementedError):
            differing = 'First %s has no length.    Non-sequence?' % (
                seq_type_name)

        if differing is None:
            try:
                len2 = len(seq2)
            except (TypeError, NotImplementedError):
                differing = 'Second %s has no length.    Non-sequence?' % (
                    seq_type_name)

        if differing is None:
            seq1_repr = unittest.util.safe_repr(seq1)
            seq2_repr = unittest.util.safe_repr(seq2)
            if len(seq1_repr) > 30:
                seq1_repr = seq1_repr[:30] + '...'
            if len(seq2_repr) > 30:
                seq2_repr = seq2_repr[:30] + '...'
            elements = (seq_type_name.capitalize(), seq1_repr, seq2_repr)
            differing = '%ss differ: %s != %s\n' % elements

            for i in range(min(len1, len2)):
                try:
                    item1 = seq1[i]
                except (TypeError, IndexError, NotImplementedError):
                    differing += ('\nUnable to index element %d of first %s\n'
                                  % (i, seq_type_name))
                    break

                try:
                    item2 = seq2[i]
                except (TypeError, IndexError, NotImplementedError):
                    differing += ('\nUnable to index element %d of second %s\n'
                                  % (i, seq_type_name))
                    break

                self.assertEqualNested(
                    item1, item2,
                    msg=differing + '\nFirst differing element at index %d' % i)
            else:
                if (len1 == len2):
                    return

            if len1 > len2:
                differing += ('\nFirst %s contains %d additional '
                              'elements.\n' % (seq_type_name, len1 - len2))
                try:
                    differing += ('First extra element %d:\n%s\n' %
                                  (len2, unittest.util.safe_repr(seq1[len2])))
                except (TypeError, IndexError, NotImplementedError):
                    differing += ('Unable to index element %d '
                                  'of first %s\n' % (len2, seq_type_name))
            elif len1 < len2:
                differing += ('\nSecond %s contains %d additional '
                              'elements.\n' % (seq_type_name, len2 - len1))
                try:
                    differing += ('First extra element %d:\n%s\n' %
                                  (len1, unittest.util.safe_repr(seq2[len1])))
                except (TypeError, IndexError, NotImplementedError):
                    differing += ('Unable to index element %d '
                                  'of second %s\n' % (len1, seq_type_name))
        standardMsg = differing
        diffMsg = '\n' + '\n'.join(
            difflib.ndiff(pprint.pformat(seq1).splitlines(),
                          pprint.pformat(seq2).splitlines()))
        standardMsg = self._truncateMessage(standardMsg, diffMsg)
        msg = self._formatMessage(msg, standardMsg)
        self.fail(msg)
        
    def assertDictEqualNested(self, d1, d2, msg=None):
        self.assertIsInstance(d1, dict, 'First argument is not a dictionary')
        self.assertIsInstance(d2, dict, 'Second argument is not a dictionary')
        self.assertEqual(sorted(list(d1.keys())), sorted(list(d2.keys())),
                         'Dictionaries do not have the same keys')
        for k in d1.keys():
            standardMsg = 'Value for key %s differs' % k
            msg_k = self._formatMessage(msg, standardMsg)
            self.assertEqual(d1[k], d2[k], msg=msg_k)

    def assertUnitsEqual(self, first, second, msg=None):
        r"""Assertion for equality in case of objects with units."""
        if units.has_units(first) and units.has_units(second):
            first = units.convert_to(first, units.get_units(second))
        self.assertEqual(units.get_data(first), units.get_data(second))
        
    def assertArrayEqual(self, first, second, msg=None):
        r"""Assertion for equality in case of arrays."""
        try:
            np.testing.assert_array_equal(first, second)
        except AssertionError as e:
            standardMsg = str(e)
            msg = self._formatMessage(msg, standardMsg)
            raise self.failureException(msg)


if backwards.PY2:  # pragma: Python 2
    # Dummy TestCase instance, so we can initialize an instance
    # and access the assert instance methods
    class DummyTestCase(WrappedTestCase):  # pragma: no cover
        def __init__(self):
            super(DummyTestCase, self).__init__('_dummy')

        def _dummy(self):
            pass

    # A metaclass that makes __getattr__ static
    class AssertsAccessorType(type):  # pragma: no cover
        dummy = DummyTestCase()

        def __getattr__(cls, key):
            return getattr(AssertsAccessor.dummy, key)

    # The actual accessor, a static class, that redirect the asserts
    class AssertsAccessor(object):  # pragma: no cover
        __metaclass__ = AssertsAccessorType
        
    ut = AssertsAccessor
        
else:  # pragma: Python 3

    ut = WrappedTestCase()


long_running = unittest.skipIf(not enable_long_tests, "Long tests not enabled.")
extra_example = unittest.skipIf(skip_extra_examples, "Extra examples not enabled.")


# def long_running(func):
#     r"""Decorator for marking long tests that should be skipped if
#     YGG_ENABLE_LONG_TESTS is set.

#     Args:
#         func (callable): Test function or method.

#     """
#     return unittest.skipIf(not enable_long_tests, "Long tests not enabled.")(func)


def assert_raises(exception, *args, **kwargs):
    r"""Assert that a call raises an exception.

    Args:
        exception (Exception): Exception class that should be raised.
        callable (function, class, optional): Callable that should raise the
            exception. If not provided, a context manager is returned.
        *args: Additional arguments are passed to the callable.
        **kwargs: Additional keyword arguments are passed to the callable.

    Raises:
        AssertionError: If the correct exception is not raised.

    """
    return ut.assertRaises(exception, *args, **kwargs)


@contextlib.contextmanager
def assert_warns(warning, *args, **kwargs):
    r"""Assert that a call (or context) raises an exception.

    Args:
        warning (Warning): Warning class that should be raised.
        callable (function, class, optional): Function that should raise
            the warning. If not provided, a context manager is returned.
        *args: Additional arguments are passed to the callable.
        **kwargs: Additional keyword arguments are passed to the callable.

    Raises:
        AssertionError: If the correct warning is not caught.

    """
    if backwards.PY2:  # pragma: Python 2
        if args and args[0] is None:  # pragma: debug
            warnings.warn("callable is None",
                          DeprecationWarning, 3)
            args = ()
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            try:
                if not args:
                    yield w
                else:  # pragma: debug
                    callable_obj = args[0]
                    args = args[1:]
                    callable_obj(*args, **kwargs)
            finally:
                assert(len(w) >= 1)
                for iw in w:
                    assert(issubclass(iw.category, warning))
    else:  # pragma: Python 3
        yield ut.assertWarns(warning, *args, **kwargs)


def assert_equal(x, y):
    r"""Assert that two messages are equivalent.

    Args:
        x (object): Python object to compare against y.
        y (object): Python object to compare against x.

    Raises:
        AssertionError: If the two messages are not equivalent.

    """
    ut.assertEqual(x, y)


def assert_not_equal(x, y):
    r"""Assert that two objects are NOT equivalent.

    Args:
        x (object): Python object to compare against y.
        y (object): Python object to compare against x.

    Raises:
        AssertionError: If the two objects are equivalent.

    """
    ut.assertNotEqual(x, y)
        

class YggTestBase(unittest.TestCase):
    r"""Wrapper for unittest.TestCase that allows use of setup and
    teardown methods along with description prefix.

    Args:
        description_prefix (str, optional): String to prepend docstring
            test message with. Default to empty string.
        skip_unittest (bool, optional): If True, the unittest parent
            class will not be initialized. Defaults to False.

    Attributes:
        uuid (str): Random unique identifier.
        attr_list (list): List of attributes that should be checked for after
            initialization.
        timeout (float): Maximum time in seconds for timeouts.
        sleeptime (float): Time in seconds that should be waited for sleeps.

    """

    attr_list = list()

    def __init__(self, *args, **kwargs):
        self._description_prefix = kwargs.pop('description_prefix',
                                              str(self.__class__).split("'")[1])
        self.uuid = str(uuid.uuid4())
        self.timeout = 10.0
        self.sleeptime = 0.01
        self.attr_list = copy.deepcopy(self.__class__.attr_list)
        self._teardown_complete = False
        self._new_default_comm = None
        self._old_default_comm = None
        self._old_loglevel = None
        self._old_encoding = None
        self.debug_flag = False
        self._first_test = True
        skip_unittest = kwargs.pop('skip_unittest', False)
        if not skip_unittest:
            super(YggTestBase, self).__init__(*args, **kwargs)

    def assert_equal(self, x, y):
        r"""Assert that two values are equal."""
        return assert_equal(x, y)

    def assert_less_equal(self, x, y):
        r"""Assert that one value is less than or equal to another."""
        return self.assertLessEqual(x, y)

    def assert_greater(self, x, y):
        r"""Assert that one value is greater than another."""
        return self.assertGreater(x, y)

    def assert_raises(self, *args, **kwargs):
        r"""Assert that a function raises an error."""
        return self.assertRaises(*args, **kwargs)

    @property
    def comm_count(self):
        r"""int: The number of comms."""
        out = 0
        for k in self.cleanup_comm_classes:
            cls = import_component('comm', k)
            out += cls.comm_count()
        return out

    @property
    def fd_count(self):
        r"""int: The number of open file descriptors."""
        proc = psutil.Process()
        if platform._is_win:  # pragma: windows
            out = proc.num_handles()
        else:
            out = proc.num_fds()
        # print(proc.num_fds(), proc.num_threads(), len(proc.connections("all")),
        #      len(proc.open_files()))
        return out

    @property
    def thread_count(self):
        r"""int: The number of active threads."""
        return threading.active_count()

    def set_utf8_encoding(self):
        r"""Set the encoding to utf-8 if it is not already."""
        old_lang = os.environ.get('LANG', '')
        if 'UTF-8' not in old_lang:  # pragma: debug
            self._old_encoding = old_lang
            os.environ['LANG'] = 'en_US.UTF-8'
            
    def reset_encoding(self):
        r"""Reset the encoding to the original value before the test."""
        if self._old_encoding is not None:  # pragma: debug
            os.environ['LANG'] = self._old_encoding
            self._old_encoding = None

    def debug_log(self):  # pragma: debug
        r"""Turn on debugging."""
        self._old_loglevel = ygg_cfg.get('debug', 'ygg')
        ygg_cfg.set('debug', 'ygg', 'DEBUG')
        cfg_logging()

    def reset_log(self):  # pragma: debug
        r"""Resetting logging to prior value."""
        if self._old_loglevel is not None:
            ygg_cfg.set('debug', 'ygg', self._old_loglevel)
            cfg_logging()
            self._old_loglevel = None

    def set_default_comm(self, default_comm=None):
        r"""Set the default comm."""
        self._old_default_comm = os.environ.get('YGG_DEFAULT_COMM', None)
        if default_comm is None:
            default_comm = self._new_default_comm
        if default_comm is not None:
            from yggdrasil.communication.DefaultComm import DefaultComm
            os.environ['YGG_DEFAULT_COMM'] = default_comm
            DefaultComm._reset_alias()

    def reset_default_comm(self):
        r"""Reset the default comm to the original value."""
        if self._old_default_comm is None:
            if 'YGG_DEFAULT_COMM' in os.environ:
                del os.environ['YGG_DEFAULT_COMM']
        else:  # pragma: debug
            os.environ['YGG_DEFAULT_COMM'] = self._old_default_comm

    def setUp(self, *args, **kwargs):
        self.setup(*args, **kwargs)

    def tearDown(self, *args, **kwargs):
        self.teardown(*args, **kwargs)

    def setup(self, nprev_comm=None, nprev_thread=None, nprev_fd=None):
        r"""Record the number of open comms, threads, and file descriptors.

        Args:
            nprev_comm (int, optional): Number of previous comm channels.
                If not provided, it is determined to be the present number of
                default comms.
            nprev_thread (int, optional): Number of previous threads.
                If not provided, it is determined to be the present number of
                threads.
            nprev_fd (int, optional): Number of previous open file descriptors.
                If not provided, it is determined to be the present number of
                open file descriptors.

        """
        self.set_default_comm()
        self.set_utf8_encoding()
        if self.debug_flag:  # pragma: debug
            self.debug_log()
        if nprev_comm is None:
            nprev_comm = self.comm_count
        if nprev_thread is None:
            nprev_thread = self.thread_count
        if nprev_fd is None:
            nprev_fd = self.fd_count
        self.nprev_comm = nprev_comm
        self.nprev_thread = nprev_thread
        self.nprev_fd = nprev_fd

    def teardown(self, ncurr_comm=None, ncurr_thread=None, ncurr_fd=None):
        r"""Check the number of open comms, threads, and file descriptors.

        Args:
            ncurr_comm (int, optional): Number of current comms. If not
                provided, it is determined to be the present number of comms.
            ncurr_thread (int, optional): Number of current threads. If not
                provided, it is determined to be the present number of threads.
            ncurr_fd (int, optional): Number of current open file descriptors.
                If not provided, it is determined to be the present number of
                open file descriptors.

        """
        self._teardown_complete = True
        x = tools.YggClass('dummy', timeout=self.timeout, sleeptime=self.sleeptime)
        # Give comms time to close
        if ncurr_comm is None:
            Tout = x.start_timeout()
            while ((not Tout.is_out)
                   and (self.comm_count > self.nprev_comm)):  # pragma: debug
                x.sleep()
            x.stop_timeout()
            ncurr_comm = self.comm_count
        self.assert_less_equal(ncurr_comm, self.nprev_comm)
        # Give threads time to close
        if ncurr_thread is None:
            Tout = x.start_timeout()
            while ((not Tout.is_out)
                   and (self.thread_count > self.nprev_thread)):  # pragma: debug
                x.sleep()
            x.stop_timeout()
            ncurr_thread = self.thread_count
        self.assert_less_equal(ncurr_thread, self.nprev_thread)
        # Give files time to close
        self.cleanup_comms()
        if ncurr_fd is None:
            if not self._first_test:
                Tout = x.start_timeout()
                while ((not Tout.is_out)
                       and (self.fd_count > self.nprev_fd)):  # pragma: debug
                    x.sleep()
                x.stop_timeout()
            ncurr_fd = self.fd_count
        fds_created = ncurr_fd - self.nprev_fd
        # print("FDS CREATED: %d" % fds_created)
        if not self._first_test:
            self.assert_equal(fds_created, 0)
        # Reset the log, encoding, and default comm
        self.reset_log()
        self.reset_encoding()
        self.reset_default_comm()
        self._first_test = False

    @property
    def cleanup_comm_classes(self):
        r"""list: Comm classes that should be cleaned up following the test."""
        return [tools.get_default_comm()]

    def cleanup_comms(self):
        r"""Cleanup all comms."""
        for k in self.cleanup_comm_classes:
            cleanup_comms(k)

    @property
    def description_prefix(self):
        r"""String prefix to prepend docstr test message with."""
        return self._description_prefix

    def shortDescription(self):
        r"""Prefix first line of doc string."""
        out = super(YggTestBase, self).shortDescription()
        if self.description_prefix:
            out = '%s: %s' % (self.description_prefix, out)
        return out

    def check_file_exists(self, fname):
        r"""Check that a file exists.

        Args:
            fname (str): Full path to the file that should be checked.

        """
        Tout = self.start_timeout(2)
        while (not Tout.is_out) and (not os.path.isfile(fname)):  # pragma: debug
            self.sleep()
        self.stop_timeout()
        if not os.path.isfile(fname):  # pragma: debug
            raise AssertionError("File '%s' dosn't exist." % fname)

    def check_file_size(self, fname, fsize):
        r"""Check that file is the correct size.

        Args:
            fname (str): Full path to the file that should be checked.
            fsize (int): Size that the file should be in bytes.

        """
        Tout = self.start_timeout(2)
        if (os.stat(fname).st_size != fsize):  # pragma: debug
            print('file sizes not equal', os.stat(fname).st_size, fsize)
        while ((not Tout.is_out)
               and (os.stat(fname).st_size != fsize)):  # pragma: debug
            self.sleep()
        self.stop_timeout()
        if os.stat(fname).st_size != fsize:  # pragma: debug
            raise AssertionError("File size (%d), dosn't match expected size (%d)."
                                 % (os.stat(fname).st_size, fsize))

    def check_file_contents(self, fname, result):
        r"""Check that the contents of a file are correct.

        Args:
            fname (str): Full path to the file that should be checked.
            result (str): Contents of the file.

        """
        with open(fname, 'r') as fd:
            ocont = fd.read()
        if ocont != result:  # pragma: debug
            odiff = '\n'.join(list(difflib.Differ().compare(ocont, result)))
            raise AssertionError(('File contents do not match expected result.'
                                  'Diff:\n%s') % odiff)

    def check_file(self, fname, result):
        r"""Check that a file exists, is the correct size, and has the correct
        contents.

        Args:
            fname (str): Full path to the file that should be checked.
            result (str): Contents of the file.

        """
        self.check_file_exists(fname)
        self.check_file_size(fname, len(result))
        self.check_file_contents(fname, result)


class YggTestClass(YggTestBase):
    r"""Test class for a YggClass."""

    testing_option_kws = {}
    _mod = None
    _cls = None
    skip_init = False

    def __init__(self, *args, **kwargs):
        self._inst_args = list()
        self._inst_kwargs = dict()
        self._extra_instances = []
        super(YggTestClass, self).__init__(*args, **kwargs)

    def setup(self, *args, **kwargs):
        r"""Create an instance of the class."""
        super(YggTestClass, self).setup(*args, **kwargs)
        if not self.skip_init:
            self._instance = self.create_instance()

    def teardown(self, *args, **kwargs):
        r"""Remove the instance."""
        self.clear_instance()
        super(YggTestClass, self).teardown(*args, **kwargs)
        for i in range(len(self._extra_instances)):
            inst = self._extra_instances[i]
            self._extra_instances[i] = None
            self.remove_instance(inst)
            del inst
        self._extra_instances = []

    @property
    def description_prefix(self):
        r"""String prefix to prepend docstr test message with."""
        if self.cls is None:
            return super(YggTestClass, self).description_prefix
        else:
            return self.cls

    @property
    def cls(self):
        r"""str: Class to be tested."""
        return self._cls

    @property
    def mod(self):
        r"""str: Absolute name of module containing class to be tested."""
        return self._mod

    @property
    def inst_args(self):
        r"""list: Arguments for creating a class instance."""
        return self._inst_args

    @property
    def inst_kwargs(self):
        r"""dict: Keyword arguments for creating a class instance."""
        out = self._inst_kwargs
        return out

    @property
    def import_cls(self):
        r"""Import the tested class from its module"""
        if self.mod is None:
            raise Exception("No module registered.")
        if self.cls is None:
            raise Exception("No class registered.")
        mod = importlib.import_module(self.mod)
        cls = getattr(mod, self.cls)
        return cls

    def get_options(self):
        r"""Get testing options."""
        if self.mod is None:  # pragma: debug
            return {}
        return self.import_cls.get_testing_options(**self.testing_option_kws)

    @property
    def testing_options(self):
        r"""dict: Testing options."""
        if getattr(self, '_testing_options', None) is None:
            self._testing_options = self.get_options()
        return self._testing_options

    @property
    def instance(self):
        r"""object: Instance of the test driver."""
        if self._teardown_complete:
            raise RuntimeError("Instance referenced after teardown.")
        if self.skip_init:  # pragma: debug
            raise RuntimeError("skip_init is True, so instance cannot be used.")
        if not hasattr(self, '_instance'):  # pragma: debug
            self._instance = self.create_instance()
        return self._instance

    def create_error_instance(self, inst_class=None, args=None, kwargs=None,
                              error_class=None, error_on_init=False):  # pragma: no cover
        r"""Create a new instance of the class that is wrapped in ErrorClass."""
        if inst_class is None:
            inst_class = self.import_cls
        if args is None:
            args = self.inst_args
        if kwargs is None:
            kwargs = self.inst_kwargs
        if error_class is None:
            error_class = ErrorClass
        if error_class == ErrorClass:
            # This could be a normal class that contains error classes
            args.insert(0, inst_class)
            kwargs['error_on_init'] = error_on_init
        error_kwargs = dict(inst_class=error_class, args=args, kwargs=kwargs)
        if error_on_init:
            self.assert_raises(MagicTestError, self.create_instance, **error_kwargs)
        else:
            out = self.create_instance(**error_kwargs)
            self._extra_instances.append(out)
            return out

    def create_instance(self, inst_class=None, args=None, kwargs=None):
        r"""Create a new instance of the class."""
        if inst_class is None:
            inst_class = self.import_cls
        if args is None:
            args = self.inst_args
        if kwargs is None:
            kwargs = self.inst_kwargs
        inst = inst_class(*args, **kwargs)
        return inst

    def remove_instance(self, inst):
        r"""Remove an instance of the class."""
        pass

    def clear_instance(self):
        r"""Clear the instance."""
        if hasattr(self, '_instance'):
            inst = self._instance
            self._instance = None
            self.remove_instance(inst)
            delattr(self, '_instance')


class IOInfo(object):
    r"""Simple class for useful IO attributes."""

    def __init__(self):
        self.field_names = ['name', 'count', 'size']
        self.field_units = ['n/a', 'umol', 'cm']
        self.nfields = len(self.field_names)
        self.comment = b'# '
        self.delimiter = b'\t'
        self.newline = b'\n'
        self.field_names = [backwards.as_bytes(x) for x in self.field_names]
        self.field_units = [backwards.as_bytes(x) for x in self.field_units]


class YggTestClassInfo(YggTestClass, IOInfo):
    r"""Test class for a YggClass with IOInfo available."""

    def __init__(self, *args, **kwargs):
        super(YggTestClassInfo, self).__init__(*args, **kwargs)
        IOInfo.__init__(self)


class MagicTestError(Exception):
    r"""Special exception for testing."""
    pass

            
def ErrorClass(base_class, *args, **kwargs):
    r"""Wrapper to return errored version of a class.

    Args:
        base_class (class): Base class to use.
        *args: Additional arguments are passed to the class constructor.
        **kwargs: Additional keyword arguments are passed to the class
            constructor.

    """

    class ErrorClass(base_class):
        r"""Dummy class that will raise an error for any requested method.

        Args:
            error_on_init (bool, optional): If True, an error will be raised
                in place of the base class's __init__ method. Defaults to False.
            *args: Additional arguments are passed to the parent class.
            **kwargs: Additional keyword arguments are passed to the parent class.

        Attributes:
            error_location (str): Name of the method/attribute that will raise
                an error.

        """

        _is_error_class = True
        
        def __init__(self, *args, **kwargs):
            error_on_init = kwargs.pop('error_on_init', False)
            if error_on_init:
                self.error_method()
            self._replaced_methods = dict()
            super(ErrorClass, self).__init__(*args, **kwargs)

        def empty_method(self, *args, **kwargs):
            r"""Method that won't do anything."""
            pass
                
        def error_method(self, *args, **kwargs):
            r"""Method that will raise a MagicTestError."""
            raise MagicTestError("This is a test error.")

        def getattr(self, attr):
            r"""Get the underlying object for an attribute name."""
            for obj in [self] + self.__class__.mro():
                if attr in obj.__dict__:
                    return obj.__dict__[attr]
            raise AttributeError  # pragma: debug

        def setattr(self, attr, value):
            r"""Set the attribute at the class level."""
            setattr(self.__class__, attr, value)

        def replace_method(self, method_name, replacement):
            r"""Temporarily replace method with another."""
            self._replaced_methods[method_name] = self.getattr(method_name)
            self.setattr(method_name, replacement)
            
        def restore_method(self, method_name):
            r"""Restore the original method."""
            self.setattr(method_name, self._replaced_methods.pop(method_name))

        def restore_all(self):
            r"""Restored all replaced methods."""
            meth_list = list(self._replaced_methods.keys())
            for k in meth_list:
                self.restore_method(k)

        def empty_replace(self, method_name, **kwargs):
            r"""Replace a method with an empty method."""
            self.replace_method(method_name, self.empty_method, **kwargs)

        def error_replace(self, method_name, **kwargs):
            r"""Replace a method with an errored method."""
            self.replace_method(method_name, self.error_method, **kwargs)

    return ErrorClass(*args, **kwargs)


__all__ = ['data', 'scripts', 'yamls', 'IOInfo', 'ErrorClass',
           'YggTestBase', 'YggTestClass',
           'YggTestBaseInfo', 'YggTestClassInfo']



from yggdrasil.drivers.ServerDriver import ServerDriver


class RMQAsyncServerDriver(ServerDriver):
    r"""Class for handling an RMQAsync server.

    Args:
        *args: Arguments are passed to parent class.
        **kwargs: Keyword arguments are passed to parent class.

    """

    _connection_type = 'rmq_async_server'
    _schema_subtype_description = ('Connection between an asynchronous RabbitMQ '
                                   'server request comm and a model acting as a '
                                   'server.')
    
    def __init__(self, *args, **kwargs):
        kwargs['comm'] = 'RMQAsyncComm'
        super(RMQAsyncServerDriver, self).__init__(*args, **kwargs)



import logging
from logging import debug
import os
import sys
from YggInterface import YggInput, YggOutput


def runhello():
    debug('hello pythonfiles from %s', os.getcwd())
    inf = YggInput('inFile')
    outf = YggOutput('outFile')
    fdata = inf.recv()
    debug('got input %s', fdata)
    print(fdata)
    outf.send(fdata)
    debug('sent output')
    debug('bye')


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG,
                        format=sys.argv[0].split('/')[-1] + ':%(message)s')
    runhello()



import copy
import jsonschema
from yggdrasil.metaschema.datatypes import get_type_class, _type_registry
from yggdrasil.metaschema.properties import get_metaschema_property
from yggdrasil.metaschema.datatypes.JSONObjectMetaschemaType import (
    JSONObjectMetaschemaType)


def _specificity_sort_key(item):
    return item.specificity


def _normalize_schema(validator, ref, instance, schema):
    r"""Normalize a schema at the root to handle case where only type
    string specified."""
    # if isinstance(instance, str):
    #     instance = dict(type=instance)
    # return instance
    if isinstance(instance, str) and (instance in _type_registry):
        instance = {'type': instance}
    elif isinstance(instance, dict):
        if len(instance) == 0:
            pass
        elif 'type' not in instance:
            valid_types = None
            for k in instance.keys():
                prop_class = get_metaschema_property(k, skip_generic=True)
                if prop_class is None:
                    continue
                if valid_types is None:
                    valid_types = set(prop_class.types)
                else:
                    valid_types = (valid_types & set(prop_class.types))
            if (valid_types is None) or (len(valid_types) == 0):
                # There were not any recorded properties so this must be a
                # dictionary of properties
                instance = {'type': 'object', 'properties': instance}
            else:
                if len(valid_types) > 1:
                    valid_type_classes = sorted([_type_registry[t] for t in valid_types],
                                                key=_specificity_sort_key)
                    s_max = valid_type_classes[0].specificity
                    valid_types = []
                    for tcls in valid_type_classes:
                        if tcls.specificity > s_max:
                            break
                        valid_types.append(tcls.name)
                    if 'scalar' in valid_types:
                        for t in ['1darray', 'ndarray']:
                            if t in valid_types:
                                valid_types.remove(t)
                    if len(valid_types) > 1:
                        raise Exception("Multiple possible classes: %s" % valid_types)
                instance['type'] = valid_types[0]
    elif isinstance(instance, (list, tuple)):
        # If inside validation of items as a schema, don't assume a
        # list is a malformed schema. Doing so results in infinite
        # recursion.
        if not ((len(validator._schema_path_stack) >= 2)
                and (validator._schema_path_stack[-2:] == ['items', 0])):
            instance = {'type': 'array', 'items': instance}
    if isinstance(instance, dict) and ('type' in instance):
        typecls = get_type_class(instance['type'])
        instance = typecls.normalize_definition(instance)
    return instance


def _validate_schema(validator, ref, instance, schema):
    r"""Validate a schema at the root to handle case where only type
    string specified."""
    if validator._normalizing and (ref == '#'):
        validator._normalized = _normalize_schema(validator, ref, instance, schema)
    errors = validator._base_validator.VALIDATORS['$ref'](
        validator, ref, instance, schema) or ()
    for e in errors:
        yield e
    if validator._normalizing and (ref == '#'):
        instance = validator._normalizing

        
class SchemaMetaschemaType(JSONObjectMetaschemaType):
    r"""Schema type."""

    name = 'schema'
    description = 'A schema type for evaluating subschema.'
    properties = ['type']
    definition_properties = ['type']
    metadata_properties = ['type']
    specificity = JSONObjectMetaschemaType.specificity + 1
    inherit_properties = ['extract_properties']
    _replaces_existing = False

    @classmethod
    def encode_data(cls, obj, typedef):
        r"""Encode an object's data.

        Args:
            obj (object): Object to encode.
            typedef (dict): Type definition that should be used to encode the
                object.

        Returns:
            string: Encoded object.

        """
        # Schemas should already be in JSON serializable format
        return obj

    @classmethod
    def decode_data(cls, obj, typedef):
        r"""Decode an object.

        Args:
            obj (string): Encoded object to decode.
            typedef (dict): Type definition that should be used to decode the
                object.

        Returns:
            object: Decoded object.

        """
        return obj

    @classmethod
    def validate(cls, obj, raise_errors=False):
        r"""Validate an object to check if it could be of this type.

        Args:
            obj (object): Object to validate.
            raise_errors (bool, optional): If True, errors will be raised when
                the object fails to be validated. Defaults to False.

        Returns:
            bool: True if the object could be of this type, False otherwise.

        """
        if not super(SchemaMetaschemaType, cls).validate(obj,
                                                         raise_errors=raise_errors):
            return False
        try:
            x = copy.deepcopy(cls.metaschema())
            x.setdefault('required', [])
            if 'type' not in x['required']:
                x['required'].append('type')
            x['additionalProperties'] = False
            jsonschema.validate(obj, x, cls=cls.validator())
        except jsonschema.exceptions.ValidationError:
            if raise_errors:
                raise
            return False
        return True

    @classmethod
    def normalize(cls, obj):
        r"""Normalize an object, if possible, to conform to this type.

        Args:
            obj (object): Object to normalize.

        Returns:
            object: Normalized object.

        """
        if isinstance(obj, str):
            obj = {'type': obj}
        x = cls.metaschema()
        validators = {u'$ref': _validate_schema}
        normalizers = {tuple(): [_normalize_schema]}
        validator_class = copy.deepcopy(cls.validator())
        obj = validator_class(x).normalize(obj, no_defaults=True,
                                           normalizers=normalizers,
                                           validators=validators)
        return obj



from yggdrasil.communication import CommBase, AsyncComm
from yggdrasil.config import ygg_cfg
from yggdrasil import backwards
import logging
logger = logging.getLogger(__name__)
try:
    import pika
    _rmq_installed = True
except ImportError:
    logger.warning("Could not import pika. "
                   + "RabbitMQ support will be disabled.")
    pika = None
    _rmq_installed = False


_rmq_param_sep = '_RMQPARAM_'
if backwards.PY2:  # pragma: Python 2
    
    class BlockingIOError:
        pass


def check_rmq_server(url=None, **kwargs):
    r"""Check that connection to a RabbitMQ server is possible.

    Args:
        url (str, optional): Address of RMQ server that includes all the
            necessary information. If this is provided, the remaining arguments
            are ignored. Defaults to None and the connection parameters are
            taken from the other arguments.
        username (str, optional): RMQ server username. Defaults to config value.
        password (str, optional): RMQ server password. Defaults to config value.
        host (str, optional): RMQ hostname or IP Address to connect to. Defaults
            to config value.
        port (str, optional): RMQ server TCP port to connect to. Defaults to
            config value.
        vhost (str, optional): RMQ virtual host to use. Defaults to config value.

    Returns:
        bool: True if connection to RabbitMQ server is possible, False
            otherwise.

    """
    out = True
    if not _rmq_installed:
        return False
    if url is not None:
        parameters = pika.URLParameters(url)
    else:
        username = kwargs.get('username', ygg_cfg.get('rmq', 'user', 'guest'))
        password = kwargs.get('password', ygg_cfg.get('rmq', 'password', 'guest'))
        host = kwargs.get('host', ygg_cfg.get('rmq', 'host', 'localhost'))
        port = kwargs.get('port', ygg_cfg.get('rmq', 'port', '5672'))
        vhost = kwargs.get('vhost', ygg_cfg.get('rmq', 'vhost', '/'))
        credentials = pika.PlainCredentials(username, password)
        parameters = pika.ConnectionParameters(host=host, port=int(port),
                                               virtual_host=vhost,
                                               credentials=credentials)
    # Try to establish connection
    logging.getLogger("pika").propagate = False
    try:
        connection = pika.BlockingConnection(parameters)
        if not connection.is_open:  # pragma: debug
            raise BaseException("Connection was not openned.")
        connection.close()
    except BaseException:  # pragma: debug
        out = False
    logging.getLogger("pika").propagate = True
    return out


class RMQServer(CommBase.CommServer):
    r"""RMQ server object for cleaning up server connections."""

    def terminate(self, *args, **kwargs):
        CommBase.unregister_comm('RMQComm', self.srv_address)
        super(RMQServer, self).terminate(*args, **kwargs)


class RMQComm(AsyncComm.AsyncComm):
    r"""Class for handling basic RabbitMQ communications.

    Attributes:
        connection (:class:`pika.Connection`): RabbitMQ connection.
        channel (:class:`pika.Channel`): RabbitMQ channel.

    Raises:
        RuntimeError: If a connection cannot be established.

    Developer Notes:
        It is not advised that new language implement a RabbitMQ communication
        interface. Rather RMQ communication is included explicitly for
        connections between models that are not co-located on the same machine
        and are used by the |yggdrasil| framework connections on the Python side.

    """

    _commtype = 'rmq'
    _schema_subtype_description = ('RabbitMQ connection.')
    # Based on limit of 32bit int, this could be 2**30, but this is
    # too large for stack allocation in C so 2**20 will be used.
    _maxMsgSize = 2**20
    address_description = ("AMPQ queue address of the form "
                           "``<url>_RMQPARAM_<exchange>_RMQPARAM_<queue>`` "
                           "where ``url`` is the broker address (see explanation "
                           "`here <https://pika.readthedocs.io/en/stable/"
                           "examples/using_urlparameters.html>`_), "
                           "``exchange`` is the name of the exchange on the queue "
                           "that should be used, and ``queue`` is the name of "
                           "the queue.")
    
    def _init_before_open(self, **kwargs):
        r"""Set null connection and channel."""
        self.connection = None
        self.channel = None
        self._is_open = False
        self._bound = False
        self._server_class = RMQServer
        super(RMQComm, self)._init_before_open(**kwargs)

    @property
    def url(self):
        r"""str: AMQP server address."""
        return self.address.split(_rmq_param_sep)[0]

    @property
    def exchange(self):
        r"""str: AMQP exchange."""
        return self.address.split(_rmq_param_sep)[1]

    @property
    def queue(self):
        r"""str: AMQP queue."""
        return self.address.split(_rmq_param_sep)[2]

    @classmethod
    def new_comm_kwargs(cls, name, user=None, password=None, host=None,
                        virtual_host=None, port=None, exchange=None, queue='',
                        **kwargs):
        r"""Initialize communication with new connection.

        Args:
            name (str): Name of new connection.
            user (str, optional): RabbitMQ server username. Defaults to config
                option 'user' in section 'rmq' if it exists and 'guest' if it
                does not.
            password (str, optional): RabbitMQ server password. Defaults to
                config option 'password' in section 'rmq' if it exists and
                'guest' if it does not.
            host (str, optional): RabbitMQ server host. Defaults to config option
                'host' in section 'rmq' if it exists and 'localhost' if it
                does not. If 'localhost', the output of socket.gethostname()
                is used.
            virtual_host (str, optional): RabbitMQ server virtual host. Defaults
                to config option 'vhost' in section 'rmq' if it exists and '/'
                if it does not.
            port (str, optional): Port on host to use. Defaults to config option
                'port' in section 'rmq' if it exists and '5672' if it does not.
            exchange (str, optional): RabbitMQ exchange. Defaults to config
                option 'namespace' in section 'rmq' if it exits and '' if it does
                not.
            queue (str, optional): Name of the queue that messages will be
                send to or received from. If an empty string, the queue will
                be a random string and exclusive to a receiving comm. Defaults
                to ''.
            **kwargs: Additional keywords arguments are returned as keyword
                arguments for the new comm.

        Returns:
            tuple(tuple, dict): Arguments and keyword arguments for new comm.
        
        """
        args = [name]
        if 'address' not in kwargs:
            if user is None:
                user = ygg_cfg.get('rmq', 'user', 'guest')
            if password is None:
                password = ygg_cfg.get('rmq', 'password', 'guest')
            if host is None:
                host = ygg_cfg.get('rmq', 'host', 'localhost')
            # if host == 'localhost':
            #     host = socket.gethostname()
            if virtual_host is None:
                virtual_host = ygg_cfg.get('rmq', 'vhost', '/')
            if virtual_host == '/':
                virtual_host = '%2f'
            if port is None:
                port = ygg_cfg.get('rmq', 'port', '5672')
            if exchange is None:
                exchange = ygg_cfg.get('rmq', 'namespace', '')
            url = 'amqp://%s:%s@%s:%s/%s' % (
                user, password, host, port, virtual_host)
            kwargs['address'] = _rmq_param_sep.join([url, exchange, queue])
        return args, kwargs

    @classmethod
    def underlying_comm_class(self):
        r"""str: Name of underlying communication class."""
        return 'RMQComm'

    def opp_comm_kwargs(self):
        r"""Get keyword arguments to initialize communication with opposite
        comm object.

        Returns:
            dict: Keyword arguments for opposite comm object.

        """
        kwargs = super(RMQComm, self).opp_comm_kwargs()
        return kwargs

    def bind(self):
        r"""Declare queue to get random new queue."""
        if self.is_open or self._bound:
            return
        self._bound = True
        parameters = pika.URLParameters(self.url)
        self.connection = pika.BlockingConnection(parameters)
        self.channel = self.connection.channel()
        self.channel.exchange_declare(exchange=self.exchange,
                                      auto_delete=True)
        if self.direction == 'recv' and not self.queue:
            exclusive = False  # True
        else:
            exclusive = False
        if self.queue.startswith('amq.'):
            passive = True
        else:
            passive = False
        res = self.channel.queue_declare(queue=self.queue,
                                         exclusive=exclusive,
                                         passive=passive,
                                         auto_delete=True)
        if not self.queue:
            self.address += res.method.queue
        self.channel.queue_bind(exchange=self.exchange,
                                # routing_key=self.routing_key,
                                queue=self.queue)
        self.register_comm(self.address, (self.connection, self.channel))
        super(RMQComm, self).bind()

    def _open_direct(self):
        r"""Open connection and bind/connect to queue as necessary."""
        super(RMQComm, self)._open_direct()
        if not self.is_open:
            self.bind()
            self._is_open = True
            self._bound = False

    def _close_direct(self, linger=False):
        r"""Close the connection.

        Args:
            linger (bool, optional): If True, drain messages before closing the
                comm. Defaults to False.

        """
        self._is_open = False
        self._bound = False
        if self.direction == 'recv':
            self.close_queue()
        self.close_channel()
        self.close_connection()
        if not self.is_client:
            self.unregister_comm(self.address)
        self.connection = None
        self.channel = None
        super(RMQComm, self)._close_direct(linger=linger)

    def close_queue(self):
        r"""Close the queue if the channel exists."""
        if self.channel and (not self.is_client):
            try:
                self.channel.queue_unbind(queue=self.queue,
                                          exchange=self.exchange)
                self.channel.queue_delete(queue=self.queue)
            except (pika.exceptions.ChannelClosed,
                    pika.exceptions.ConnectionClosed):  # pragma: debug
                pass
            except AttributeError:  # pragma: debug
                if self.channel is not None:
                    raise

    def close_channel(self):
        r"""Close the channel if it exists."""
        if self.channel:
            try:
                self.channel.close()
            except (pika.exceptions.ChannelClosed,
                    pika.exceptions.ConnectionClosed,
                    pika.exceptions.ChannelAlreadyClosing):  # pragma: debug
                pass
        self.channel = None

    def close_connection(self):
        r"""Close the connection."""
        if self.connection:
            try:
                self.connection.close()
            except (pika.exceptions.ChannelClosed,
                    pika.exceptions.ConnectionClosed,
                    pika.exceptions.ChannelAlreadyClosing):  # pragma: debug
                pass
            except AttributeError:  # pragma: debug
                pass
        self.connection = None

    @property
    def is_open_direct(self):
        r"""bool: True if the connection and channel are open."""
        # with self._closing_thread.lock:
        if self.channel is None or self.connection is None:
            return False
        if self.connection.is_open:
            if self.connection.is_closing:  # pragma: debug
                return False
        else:  # pragma: debug
            return False
        if self.channel.is_open:
            if self.channel.is_closing:  # pragma: debug
                return False
        else:  # pragma: debug
            return False
        return self._is_open

    def get_queue_result(self):
        r"""Get the fram from passive queue declare."""
        res = None
        if self.is_open_direct:
            try:
                res = self.channel.queue_declare(queue=self.queue,
                                                 auto_delete=True,
                                                 passive=True)
            except BlockingIOError:  # pragma: debug
                self.sleep()
                res = self.get_queue_result()
            except (pika.exceptions.ChannelClosed,
                    pika.exceptions.ConnectionClosed,
                    AttributeError):  # pragma: debug
                self._close_direct()
        return res
        
    @property
    def n_msg_direct_recv(self):
        r"""int: Number of messages in the queue."""
        return self.n_msg_direct_send

    @property
    def n_msg_direct_send(self):
        r"""int: Number of messages in the queue."""
        out = 0
        # with self._closing_thread.lock:
        if self.is_open_direct:
            res = self.get_queue_result()
            if res is not None:
                out = res.method.message_count
        return out

    @property
    def is_confirmed_send(self):
        r"""bool: True if all sent messages have been confirmed."""
        return True

    @property
    def is_confirmed_recv(self):
        r"""bool: True if all received messages have been confirmed."""
        return True

    @property
    def get_work_comm_kwargs(self):
        r"""dict: Keyword arguments for an existing work comm."""
        out = super(RMQComm, self).get_work_comm_kwargs
        out['exchange'] = self.exchange
        return out

    @property
    def create_work_comm_kwargs(self):
        r"""dict: Keyword arguments for a new work comm."""
        out = super(RMQComm, self).create_work_comm_kwargs
        out['exchange'] = self.exchange
        return out

    def _send_multipart_worker(self, msg, header, **kwargs):
        r"""Send multipart message to the worker comm identified.

        Args:
            msg (str): Message to be sent.
            header (dict): Message info including work comm address.

        Returns:
            bool: Success or failure of sending the message.

        """
        self.sched_task(0.0, super(RMQComm, self)._send_multipart_worker,
                        args=[msg, header], kwargs=kwargs, store_output=True)
        T = self.start_timeout()
        while (not T.is_out) and (self.sched_out is None):
            self.sleep()
        self.stop_timeout()
        out = self.sched_out
        self.sched_out = None
        # workcomm = self.get_work_comm(header)
        # args = [msg]
        # self.sched_task(self.sleeptime, workcomm._send_multipart,
        #                 args=args, kwargs=kwargs)
        # self.sched_task(1, self.remove_work_comm,
        #                 args=[header['id']], kwargs=dict(dont_close=True))
        # self.remove_work_comm(header['id'], dont_close=True)
        return out
    
    def _send_direct(self, msg, exchange=None, routing_key=None, **kwargs):
        r"""Send a message.

        Args:
            msg (str, bytes): Message to be sent.
            exchange (str, optional): Exchange that message should be sent
                to. Defaults to self.exchange.
            routing_key (str, optional): Key that exchange should use to route
                the message. Defaults to self.queue.
            **kwargs: Additional keyword arguments are passed to
                :method:`pika.BlockingChannel.basic_publish`.

        Returns:
            bool: Success or failure of send.

        """
        if exchange is None:
            exchange = self.exchange
        if routing_key is None:
            routing_key = self.queue
        kwargs.setdefault('mandatory', True)
        out = self.channel.basic_publish(exchange, routing_key, msg, **kwargs)
        return out

    def _recv_direct(self):
        r"""Receive a message.

        Returns:
            tuple (bool, obj): Success or failure of receive and received
                message.

        """
        method_frame, props, msg = self.channel.basic_get(
            queue=self.queue, no_ack=False)
        if method_frame:
            self.channel.basic_ack(method_frame.delivery_tag)
        else:  # pragma: debug
            self.debug("No message")
            msg = self.empty_bytes_msg
        return (True, msg)

    def purge(self):
        r"""Remove all messages from the associated queue."""
        with self._closing_thread.lock:
            if self.channel:
                self.channel.queue_purge(queue=self.queue)
        super(RMQComm, self).purge()



from yggdrasil import units
from yggdrasil.metaschema.datatypes.ScalarMetaschemaType import (
    ScalarMetaschemaType)
from yggdrasil.metaschema.properties import ScalarMetaschemaProperties


class OneDArrayMetaschemaType(ScalarMetaschemaType):
    r"""Type associated with a scalar."""

    name = '1darray'
    description = 'A 1D array with or without units.'
    properties = ['length']
    metadata_properties = ['length']
    python_types = ScalarMetaschemaProperties._all_python_arrays

    @classmethod
    def validate(cls, obj, raise_errors=False):
        r"""Validate an object to check if it could be of this type.

        Args:
            obj (object): Object to validate.
            raise_errors (bool, optional): If True, errors will be raised when
                the object fails to be validated. Defaults to False.

        Returns:
            bool: True if the object could be of this type, False otherwise.

        """
        if not super(OneDArrayMetaschemaType, cls).validate(obj,
                                                            raise_errors=raise_errors):
            return False
        if units.get_data(obj).ndim != 1:
            if raise_errors:
                raise ValueError("The array has more than one dimension.")
            return False
        return True


class NDArrayMetaschemaType(ScalarMetaschemaType):
    r"""Type associated with a scalar."""

    name = 'ndarray'
    description = 'An ND array with or without units.'
    properties = ['shape']
    metadata_properties = ['shape']
    python_types = ScalarMetaschemaProperties._all_python_arrays

    @classmethod
    def validate(cls, obj, raise_errors=False):
        r"""Validate an object to check if it could be of this type.

        Args:
            obj (object): Object to validate.
            raise_errors (bool, optional): If True, errors will be raised when
                the object fails to be validated. Defaults to False.

        Returns:
            bool: True if the object could be of this type, False otherwise.

        """
        if not super(NDArrayMetaschemaType, cls).validate(obj,
                                                          raise_errors=raise_errors):
            return False
        if units.get_data(obj).ndim <= 1:
            if raise_errors:
                raise ValueError("The array does not have more than one dimension.")
            return False
        return True



from yggdrasil.metaschema.datatypes import (
    get_type_class, complete_typedef, encode_data, encode_data_readable)
from yggdrasil.metaschema.datatypes.MetaschemaType import MetaschemaType


class ContainerMetaschemaType(MetaschemaType):
    r"""Type associated with a container of subtypes."""

    name = 'container'
    description = 'A container of other types.'
    python_types = []

    _container_type = None
    _json_type = None
    _json_property = None

    def __init__(self, *args, **kwargs):
        self._typecls = self._container_type()
        super(ContainerMetaschemaType, self).__init__(*args, **kwargs)

    @classmethod
    def _iterate(cls, container):
        r"""Iterate over the contents of the container. Each element returned
        should be a tuple including an index and a value.

        Args:
            container (obj): Object to be iterated over.

        Returns:
            iterator: Iterator over elements in the container.

        """
        raise NotImplementedError("This must be overwritten by the subclass.")

    @classmethod
    def _assign(cls, container, index, value):
        r"""Assign an element in the container to the specified value.

        Args:
            container (obj): Object that element will be assigned to.
            index (obj): Index in the container object where element will be
                assigned.
            value (obj): Value that will be assigned to the element in the
                container object.

        """
        raise NotImplementedError("This must be overwritten by the subclass.")

    @classmethod
    def _has_element(cls, container, index):
        r"""Check to see if an index is in the container.

        Args:
            container (obj): Object that should be checked for index.
            index (obj): Index that should be checked for.

        Returns:
            bool: True if the index is in the container.

        """
        raise NotImplementedError("This must be overwritten by the subclass.")

    @classmethod
    def _get_element(cls, container, index, default):
        r"""Get an element from the container if it exists, otherwise return
        the default.

        Args:
            container (obj): Object that should be returned from.
            index (obj): Index of element that should be returned.
            default (obj): Default that should be returned if the index is not
                in the container.

        Returns:
            object: Container contents at specified element.

        """
        out = default
        if cls._has_element(container, index):
            out = container[index]
        return out

    @classmethod
    def _encode_data_alias(cls, obj, typedef, func_encode, container_type=None):
        r"""Encode an object's data using a sepcified function.

        Args:
            obj (object): Object to encode.
            typedef (dict): Type definition that should be used to encode the
                object.
            func_encode (callable): Function that should be used to encode
                elements in the container. Defaults to encode_data.
            container_type (type, optional): Type that should be used for the
                output container. Defaults to cls._container_type.

        Returns:
            string: Encoded object.

        """
        if container_type is None:
            container = cls._container_type()
        else:
            container = container_type()
        vtypedef_avail = False
        if isinstance(typedef, dict) and (cls._json_property in typedef):
            vtypedef_avail = typedef[cls._json_property]
        for k, v in cls._iterate(obj):
            vtypedef = None
            if vtypedef_avail:
                vtypedef = cls._get_element(vtypedef_avail, k, None)
            vbytes = func_encode(v, typedef=vtypedef)
            cls._assign(container, k, vbytes)
        return container

    @classmethod
    def encode_data(cls, obj, typedef, **kwargs):
        r"""Encode an object's data.

        Args:
            obj (object): Object to encode.
            typedef (dict): Type definition that should be used to encode the
                object.
            **kwargs: Additional keyword arguments are passed to the class
                method _encode_data_alias.

        Returns:
            string: Encoded object.

        """
        return cls._encode_data_alias(obj, typedef, encode_data, **kwargs)

    @classmethod
    def encode_data_readable(cls, obj, typedef, **kwargs):
        r"""Encode an object's data in a readable format.

        Args:
            obj (object): Object to encode.
            typedef (dict): Type definition that should be used to encode the
                object.
            **kwargs: Additional keyword arguments are passed to the class
                method _encode_data_alias.

        Returns:
            string: Encoded object.

        """
        return cls._encode_data_alias(obj, typedef, encode_data_readable, **kwargs)
    
    @classmethod
    def decode_data(cls, obj, typedef):
        r"""Decode an object.

        Args:
            obj (string): Encoded object to decode.
            typedef (dict): Type definition that should be used to decode the
                object.

        Returns:
            object: Decoded object.

        """
        container = cls._container_type()
        for k, v in cls._iterate(obj):
            vtypedef = cls._get_element(typedef[cls._json_property], k, {})
            vcls = get_type_class(vtypedef['type'])
            cls._assign(container, k, vcls.decode_data(v, vtypedef))
        return container

    @classmethod
    def extract_typedef(cls, metadata):
        r"""Extract the minimum typedef required for this type from the provided
        metadata.

        Args:
            metadata (dict): Message metadata.

        Returns:
            dict: Encoded type definition with unncessary properties removed.

        """
        out = super(ContainerMetaschemaType, cls).extract_typedef(metadata)
        if cls._json_property in out:
            contents = out[cls._json_property]
            if isinstance(contents, cls.python_types):
                for k, v in cls._iterate(contents):
                    if 'type' in v:
                        vcls = get_type_class(v['type'])
                        cls._assign(contents, k, vcls.extract_typedef(v))
                out[cls._json_property] = contents
        return out

    def update_typedef(self, **kwargs):
        r"""Update the current typedef with new values.

        Args:
            **kwargs: All keyword arguments are considered to be new type
                definitions. If they are a valid definition property, they
                will be copied to the typedef associated with the instance.

        Returns:
            dict: A dictionary of keyword arguments that were not added to the
                type definition.

        """
        map = kwargs.get(self._json_property, None)
        map_out = self._container_type()
        if isinstance(map, self.python_types):
            for k, v in self._iterate(map):
                v_typedef = complete_typedef(v)
                if self._has_element(self._typecls, k):
                    self._assign(map_out, k,
                                 self._typecls[k].update_typedef(**v_typedef))
                else:
                    self._assign(self._typecls, k,
                                 get_type_class(v_typedef['type'])(**v_typedef))
                self._assign(map, k, self._typecls[k]._typedef)
            kwargs[self._json_property] = map
        out = super(ContainerMetaschemaType, self).update_typedef(**kwargs)
        if map_out:
            out[self._json_property] = map_out
        return out



import unittest
import copy
from yggdrasil.tests import assert_raises, assert_equal
from yggdrasil.communication import new_comm
from yggdrasil.communication import IPCComm, CommBase
from yggdrasil.communication.tests import test_AsyncComm


_ipc_installed = IPCComm.IPCComm.is_installed(language='python')


@unittest.skipIf(not _ipc_installed, "IPC library not installed")
def test_queue():
    r"""Test creation/removal of queue."""
    mq = IPCComm.get_queue()
    key = str(mq.key)
    assert(CommBase.is_registered('IPCComm', key))
    CommBase.unregister_comm('IPCComm', key, dont_close=True)
    assert_raises(KeyError, IPCComm.remove_queue, mq)
    CommBase.register_comm('IPCComm', key, mq)
    IPCComm.remove_queue(mq)
    assert(not CommBase.is_registered('IPCComm', key))


@unittest.skipIf(not _ipc_installed, "IPC library not installed")
def test_ipcs():
    r"""Test list of ipc objects."""
    IPCComm.ipcs()

    
@unittest.skipIf(not _ipc_installed, "IPC library not installed")
def test_ipc_queues():
    r"""Test list of ipc queues."""
    IPCComm.ipc_queues()


@unittest.skipIf(not _ipc_installed, "IPC library not installed")
def test_ipcrm():
    r"""Test removal of ipc objects."""
    IPCComm.ipcrm()


@unittest.skipIf(not _ipc_installed, "IPC library not installed")
def test_ipcrm_queues():
    r"""Test removal of ipc queues."""
    IPCComm.ipcrm_queues()
    assert_equal(len(IPCComm.ipc_queues()), 0)
    mq = IPCComm.get_queue()
    assert_equal(len(IPCComm.ipc_queues()), 1)
    IPCComm.ipcrm_queues(str(mq.key))
    assert_equal(len(IPCComm.ipc_queues()), 0)

    
@unittest.skipIf(not _ipc_installed, "IPC library not installed")
class TestIPCComm(test_AsyncComm.TestAsyncComm):
    r"""Test for IPCComm communication class."""

    comm = 'IPCComm'
    attr_list = (copy.deepcopy(test_AsyncComm.TestAsyncComm.attr_list)
                 + ['q'])


@unittest.skipIf(_ipc_installed, "IPC library installed")
def test_queue_not_installed():  # pragma: windows
    r"""Test return of get_queue if IPC library is not installed."""
    assert_equal(IPCComm.get_queue(), None)


@unittest.skipIf(_ipc_installed, "IPC library installed")
def test_ipcs_not_isntalled():  # pragma: windows
    r"""Test return of ipcs if IPC library is not installed."""
    assert_equal(IPCComm.ipcs(), '')

    
@unittest.skipIf(_ipc_installed, "IPC library installed")
def test_ipcrm_not_isntalled():  # pragma: windows
    r"""Test ipcrm if IPC library is not installed."""
    IPCComm.ipcrm()

    
@unittest.skipIf(_ipc_installed, "IPC library installed")
def test_ipcrm_queues_not_isntalled():  # pragma: windows
    r"""Test ipcrm_queues if IPC library is not installed."""
    IPCComm.ipcrm_queues()

    
@unittest.skipIf(_ipc_installed, "IPC library installed")
def test_not_running():  # pragma: windows
    r"""Test raise of an error if a IPC library is not installed."""
    comm_kwargs = dict(comm='IPCComm', direction='send', reverse_names=True)
    assert_raises(RuntimeError, new_comm, 'test', **comm_kwargs)



from yggdrasil import backwards
from yggdrasil.serialize.PlySerialize import PlySerialize
from yggdrasil.metaschema.datatypes.ObjMetaschemaType import ObjDict


class ObjSerialize(PlySerialize):
    r"""Class for serializing/deserializing .obj file formats. Reader
    adapted from https://www.pygame.org/wiki/OBJFileLoader."""

    _seritype = 'obj'
    _schema_subtype_description = ('Serialize 3D structures using Obj format.')
    default_datatype = {'type': 'obj'}

    def func_serialize(self, args):
        r"""Serialize a message.

        Args:
            args: List of arguments to be formatted or numpy array to be
                serialized.

        Returns:
            bytes, str: Serialized message.

        """
        return backwards.as_bytes(self.datatype.encode_data(args, self.typedef))

    def func_deserialize(self, msg):
        r"""Deserialize a message.

        Args:
            msg: Message to be deserialized.

        Returns:
            obj: Deserialized message.

        """
        return ObjDict(self.datatype.decode_data(backwards.as_str(msg),
                                                 self.typedef))

    @classmethod
    def get_testing_options(cls):
        r"""Method to return a dictionary of testing options for this class.

        Returns:
            dict: Dictionary of variables to use for testing.

        """
        out = super(ObjSerialize, cls).get_testing_options()
        obj = ObjDict({'vertices': [{'x': float(0), 'y': float(0), 'z': float(0)},
                                    {'x': float(0), 'y': float(0), 'z': float(1)},
                                    {'x': float(0), 'y': float(1), 'z': float(1)}],
                       'faces': [[{'vertex_index': int(0)},
                                  {'vertex_index': int(1)},
                                  {'vertex_index': int(2)}]]})
        out['objects'] = [obj, obj]
        out['contents'] = (b'# Author ygg_auto\n'
                           + b'# Generated by yggdrasil\n\n'
                           + b'v 0.0000 0.0000 0.0000\n'
                           + b'v 0.0000 0.0000 1.0000\n'
                           + b'v 0.0000 1.0000 1.0000\n'
                           + b'v 0.0000 0.0000 0.0000\n'
                           + b'v 0.0000 0.0000 1.0000\n'
                           + b'v 0.0000 1.0000 1.0000\n'
                           + b'f 1// 2// 3//\n'
                           + b'f 4// 5// 6//\n')
        return out



from yggdrasil.drivers.InputDriver import InputDriver


class FileInputDriver(InputDriver):
    r"""Class that sends messages read from a file.

    Args:
        name (str): Name of the queue that messages should be sent to.
        args (str): Path to the file that messages should be read from.
        **kwargs: Additional keyword arguments are passed to the parent class.

    """

    _connection_type = 'file_input'
    _icomm_type = 'FileComm'
    _schema_subtype_description = ('Connection between a file and a model.')

    def __init__(self, name, args, **kwargs):
        kwargs.setdefault('icomm_kws', {})
        kwargs['icomm_kws']['address'] = args
        super(FileInputDriver, self).__init__(name, args, **kwargs)
        self.debug('(%s)', args)
